/* Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */


/* JSON Function items used by mysql */

#include "item_json_func.h"
#include "item_cmpfunc.h"
#include "sql_class.h"
#include "json_atom.h"          // Json_scalar_holder
#include "json_binary.h"
#include "json_dom.h"
#include "json_path.h"
#include "sql_time.h"
#include "template_utils.h"     // down_cast
#include "rapidjson/document.h"

#include <boost/variant/polymorphic_get.hpp>    // boost::polymorphic_get

using namespace json_binary;

/** For use by JSN_CONTAINS_PATH() and JSN_SEARCH() */
enum enum_one_or_all_type
{
  ooa_one,
  ooa_all,
  ooa_error
};

bool ensure_utf8mb4(String *val, String *buf,
                    const char **resptr, size_t *reslength, bool require_string)
{
  const CHARSET_INFO *cs= val->charset();

  if (cs == &my_charset_bin)
  {
    if (require_string)
      my_error(ER_INVALID_JSON_CHARSET, MYF(0), my_charset_bin.csname);
    return false;
  }

  const char *s= val->ptr();
  size_t ss= val->length();

  if (my_charset_same(cs, &my_charset_utf8mb4_bin) ||
      my_charset_same(cs, &my_charset_utf8_bin) ||
      !std::strcmp(cs->csname, "ascii"))
  {
    /*
      Character data is directly converted to JSON if the character
      set is utf8mb4 or a subset.
    */
  }
  else
  { // If not, we convert, possibly with loss (best effort).
    uint dummy_errors;
    if (buf->copy(val->ptr(), val->length(), val->charset(),
                  &my_charset_utf8mb4_bin, &dummy_errors))
    {
      my_error(ER_OUTOFMEMORY, MYF(0), val->length()*4);
      return false;
    }
    buf->set_charset(&my_charset_utf8mb4_bin);
    s= buf->ptr();
    ss= buf->length();
  }

  *resptr= s;
  *reslength= ss;
  return true;
}


static bool parse_json(String *res,
                       Item **args,
                       uint arg_idx,
                       const char *func_name,
                       bool *need_parse,
                       Json_dom **dom,
                       bool require_str_or_json)
{
  char buff[MAX_FIELD_WIDTH];
  String utf8_res(buff, sizeof(buff), &my_charset_utf8mb4_bin);
  Item *arg_item= args[arg_idx];

  if (!res)
  {
    arg_item->null_value= true;
    return true;
  }

  const char *safep;         // contents of res, possibly converted
  size_t safe_length;        // length of safep

  if (!ensure_utf8mb4(res, &utf8_res, &safep, &safe_length,
                      require_str_or_json))
  {
    if (require_str_or_json)
      arg_item->null_value= true;
    return false;
  }

  *need_parse= true;

  if (!dom)
  {
    rapidjson::Document document;
    rapidjson::StringStreamSafe ss(safep, safe_length);

    const bool success= document.AcceptAnyRoot(true).
      ParseStream<0, rapidjson::UTF8<>, rapidjson::StringStreamSafe>(ss).
      HasParseError() ? false : true;

    if (!success && require_str_or_json)
    {
      size_t offset= document.GetErrorOffset();
      const char *msg= document.GetParseError();
      std::string value(safep, safe_length);
      my_error(ER_INVALID_JSON_TEXT_IN_PARAM, MYF(0),
               arg_idx + 1, func_name, msg, offset, value.c_str());
      arg_item->null_value= !success;
    }
    return success;
  }

  const char *parse_err;
  size_t err_offset;
  *dom= Json_dom::parse(safep, safe_length, &parse_err, &err_offset);

  if (!*dom)
  {
    std::string value(safep, safe_length);
    my_error(ER_INVALID_JSON_TEXT_IN_PARAM, MYF(0),
             arg_idx + 1, func_name, parse_err, err_offset,
             value.c_str());
    arg_item->null_value= true;
  }
  return *dom != NULL;
}

/**
  Helper method for Item_func_json_* methods. Check whether an argument
  can be converted to a utf8mb4 string.

  @param[in]  arg_item    An argument Item
  @param[out] value       Where to materialize the arg_item's string value
  @param[out] utf8_res    Buffer for use by ensure_utf8mb4.
  @param[in]  func_name   Name of the user-invoked JSN_ function
  @param[out] safep       String pointer after any relevant conversion
  @param[out] safe_length Corresponding string length

  @returns false if the Item is not a utf8mb4 string
*/
static bool get_json_string(Item *arg_item,
                            String *value,
                            String *utf8_res,
                            const char *func_name,
                            const char **safep,
                            size_t *safe_length)
{
  DBUG_ENTER("get_json_string");

  String *const res= arg_item->val_str(value);

  if (!res)
  {
    arg_item->null_value= true;
    DBUG_RETURN(false);
  }

  if (!ensure_utf8mb4(res, utf8_res, safep, safe_length,
                      true))
  {
    arg_item->null_value= true;
    DBUG_RETURN(false);
  }

  DBUG_RETURN(true);
}

/**
  Helper method for Item_func_json_* methods. Check if a JSON item or
  JSON text is valid and, for the latter, optionally construct a DOM
  tree (i.e. only if valid).

  @param[in]     args       Item_func::args alias
  @param[in]     arg_idx    Index (0-based) of argument into the args array
  @param[out]    value      Item_func_json_*::m_value alias
  @param[in]     func_name  Name of the user-invoked JSN_ function
  @param[out]    need_parse True if we needed to parse to validate
  @param[in,out] dom        If non-null, we want any text parsed DOM
                            returned at the location pointed to
  @param[in]     require_str_or_json
                            If true, generate an error if other types used
                            as input

  @returns true if the text is valid JSON (or NULL), else false
*/
static bool json_is_valid(Item **args,
                          uint arg_idx,
                          String *value,
                          const char *func_name,
                          bool *need_parse,
                          Json_dom **dom,
                          bool require_str_or_json)
{
  DBUG_ENTER("json_is_valid");

  Item *const arg_item= args[arg_idx];

  switch (arg_item->field_type())
  {
  case MYSQL_TYPE_NULL:
    DBUG_ASSERT(arg_item->null_value);
    DBUG_RETURN(true);
  case MYSQL_TYPE_JSON:
    // Set the null_value flag
    arg_item->is_null();
    DBUG_RETURN(true);
  case MYSQL_TYPE_STRING:
  case MYSQL_TYPE_VAR_STRING:
  case MYSQL_TYPE_VARCHAR:
  case MYSQL_TYPE_BLOB:
  case MYSQL_TYPE_LONG_BLOB:
  case MYSQL_TYPE_MEDIUM_BLOB:
  case MYSQL_TYPE_TINY_BLOB:
    {
      String *const res= arg_item->val_str(value);
      if (arg_item->type() == Item::FIELD_ITEM)
      {
        Item_field *fi= down_cast<Item_field *>(arg_item);
        Field *field= fi->field;
        if (field->flags & (ENUM_FLAG | SET_FLAG))
        {
          DBUG_RETURN(false);
        }
      }

      const bool success= parse_json(res, args, arg_idx, func_name,
                                     need_parse, dom, require_str_or_json);
      DBUG_RETURN(success);
    }
  default:
    if (require_str_or_json)
    {
      my_error(ER_INVALID_TYPE_FOR_JSON, MYF(0), arg_idx + 1, func_name);
      arg_item->null_value= true;
    }
    DBUG_RETURN(false);
  }
}


#define CATCH_ALL(funcname, expr) \
  catch (const std::bad_alloc &e)\
  {\
    my_error(ER_STD_BAD_ALLOC_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::domain_error &e)\
  {\
    my_error(ER_STD_DOMAIN_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::length_error &e)\
  {\
    my_error(ER_STD_LENGTH_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::invalid_argument &e)\
  {\
    my_error(ER_STD_INVALID_ARGUMENT, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::out_of_range &e)\
   {\
    my_error(ER_STD_OUT_OF_RANGE_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::overflow_error &e)\
  {\
    my_error(ER_STD_OVERFLOW_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::range_error &e)\
  {\
    my_error(ER_STD_RANGE_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::underflow_error &e)\
  {\
    my_error(ER_STD_UNDERFLOW_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::logic_error &e)\
  {\
    my_error(ER_STD_LOGIC_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::runtime_error &e)\
  {\
    my_error(ER_STD_RUNTIME_ERROR, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (const std::exception &e)\
  {\
    my_error(ER_STD_UNKNOWN_EXCEPTION, MYF(0), e.what(), (funcname));\
    expr;\
  }\
  catch (...)\
  {\
    my_error(ER_STD_UNKNOWN_EXCEPTION, MYF(0), (funcname));\
    expr;\
  }

/**
  Helper method for Item_func_json_* methods. Assumes that the caller
  has already verified that the path expression is not null. Raises an
  error if the path expression is syntactically incorrect. Raises an
  error if the path expression contains wildcard tokens but is not
  supposed to. Otherwise updates the supplied Json_path object with
  the parsed path.

  @param[in]  path_expression  A string Item to be interpreted as a path.
  @param[out] value            Holder for path string
  @param[in]  forbid_wildcards True if the path shouldn't contain * or **
  @param[out] json_path        The object that will hold the parsed path

  @returns false on success, true on error or if the path is NULL
*/
static bool parse_path(Item * path_expression, String *value,
                       my_bool forbid_wildcards, Json_path *json_path)
{
  DBUG_ENTER("parse_path");

  String *path_value= path_expression->val_str(value);

  if (!path_value)
  {
    DBUG_RETURN(true);
  }

  const char * path_chars= path_value->ptr();
  size_t path_length= path_value->length();
  char buff[STRING_BUFFER_USUAL_SIZE];
  String res(buff, sizeof(buff), &my_charset_utf8mb4_bin);

  if (!ensure_utf8mb4(path_value, &res, &path_chars, &path_length, true))
  {
    DBUG_RETURN(true);
  }

  // OK, we have a string encoded in utf-8. Does it parse?
  size_t bad_idx= 0;
  if (parse_path(false, path_length, path_chars, json_path, &bad_idx))
  {
    my_error(ER_INVALID_JSON_PATH, MYF(0), bad_idx, path_value->c_ptr_safe());
    DBUG_RETURN(true);
  }

  if (forbid_wildcards && json_path->contains_wildcard_or_ellipsis())
  {
    my_error(ER_INVALID_JSON_PATH_WILDCARD, MYF(0));
    DBUG_RETURN(true);
  }

  DBUG_RETURN(false);
}

/**
   Case-insensitive comparison of a string to a lowercase string.

  @param[in]  candidate   The string to compare case-insensitively.
  @param[in]  candidate_length   The length of the candidate string.
  @param[in]  lowercase_string The lowercase string to compare it to.
  @param[in ] lowercase_length The length of the lowercase string.

  @returns True if candidate equals lower_case_string, false otherwise.
*/
static bool insensitive_cmp(const char *candidate,
                            size_t candidate_length,
                            const char *lowercase_string,
                            size_t lowercase_length)
{
  if (candidate_length != lowercase_length)
  {
    return false;
  }

  for (size_t idx = 0; idx < lowercase_length; idx++)
  {
    if (my_tolower(&my_charset_utf8mb4_bin, *candidate++) !=
        *lowercase_string++)
    {
      return false;
    }
  }

  return true;
}


/**
   Parse a oneOrAll argument.

  @param[in]  candidate   The string to compare to "one" or "all"
  @param[in]  candidate_length   The length of the candidate string.

  @returns ooa_one, ooa_all, or ooa_error, based on the match
*/
enum_one_or_all_type parse_one_or_all(const char *candidate,
                                      size_t candidate_length)
{
    if (insensitive_cmp(candidate, candidate_length, "all", 3))
    {
      return ooa_all;
    }
    else if (insensitive_cmp(candidate, candidate_length, "one", 3))
    {
      return ooa_one;
    }
    else
    {
      return ooa_error;
    }
}


type_conversion_status
Item_json_func::save_in_field(Field *field, bool no_conversions)
{
  if (field->type() == MYSQL_TYPE_JSON)
  {
    // Store the value in the JSON binary format.
    Field_json *f= down_cast<Field_json *>(field);
    Json_wrapper wr;
    val_json(&wr);

    if (null_value)
      return set_field_to_null(field);

    field->set_notnull();
    return f->store_json(&wr);
  }
  else
  {
    // TODO Convert the JSON value to text.
    return Item_func::save_in_field(field, no_conversions);
  }
}


longlong Item_func_json_valid::val_int()
{
  DBUG_ENTER("Item_func_json_valid::val_int");
  DBUG_ASSERT(fixed == 1);
  bool result= 1;
  null_value= false;

  try {
    bool dummy;
    bool ok= json_is_valid(args, 0, &m_value, func_name(), &dummy, NULL, false);

    if (!ok)
    {
      null_value= false;
      DBUG_RETURN(0);
    }

    if (args[0]->is_null() || (args[0]->null_value))
    {
      null_value= true;
      DBUG_RETURN(0);
    }

  } CATCH_ALL("jsn_valid", result= 0)

  DBUG_RETURN(result);
}

struct Wrapper_compare
{
  Json_wrapper orig;
  bool operator()(size_t l, size_t r)
  {
    return orig[l].compare(orig[r]) < 0;
  }
};

static void sort_array(const Json_wrapper &orig,
                       std::vector<size_t, Malloc_allocator<size_t> > *v)
{

  std::vector<size_t, Malloc_allocator<size_t> >
    origv((Malloc_allocator<size_t>(key_memory_JSON)));

  for (size_t i=0; i < orig.length(); i++)
  {
    origv.push_back(i);
  }

  Wrapper_compare cmp;
  cmp.orig= orig;

  std::sort(origv.begin(), origv.end(), cmp);

  // remove duplicates
  for (size_t i=0; i < orig.length(); i++)
  {
    size_t new_idx= origv[i];
    if (i > 0)
    {
      if (orig[origv[i - 1]].compare(orig[new_idx]))
      {
        v->push_back(new_idx);
      } // else equal, so don't insert
    }
    else
    {
      v->push_back(new_idx);
    }
  }
}

static bool contains_wr(const Json_wrapper &doc_wrapper,
                        const Json_wrapper &containee_wr)
{
  if (doc_wrapper.type() == Json_dom::J_OBJECT)
  {
    if (containee_wr.type() != Json_dom::J_OBJECT)
    {
      return false;
    }

    Json_wrapper_object_iterator d_oi= doc_wrapper.object_iterator();
    Json_wrapper_object_iterator c_oi= containee_wr.object_iterator();
    Json_key_comparator cmp;

    while (!c_oi.empty() && !d_oi.empty())
    {
      for( ; !d_oi.empty() && cmp(d_oi.elt().first, c_oi.elt().first);
          d_oi.next()) {}

      if (d_oi.empty() || cmp(c_oi.elt().first, d_oi.elt().first))
      {
        return false;
      }

      // key is the same, now compare values
      if (!contains_wr(d_oi.elt().second, c_oi.elt().second))
      {
        return false;
      }
      c_oi.next();
    }
    return c_oi.empty(); // must be exhausted
  }
  else if (doc_wrapper.type() == Json_dom::J_ARRAY)
  {
    const Json_wrapper *wr= &containee_wr;
    Json_wrapper a_wr;

    if (containee_wr.type() != Json_dom::J_ARRAY)
    {
      // auto-wrap scalar or object in an array for uniform treatment later
      Json_wrapper scalar= containee_wr;
      Json_array *a= new Json_array(scalar.to_dom()->clone());
      Json_wrapper nw(a);
      a_wr.steal(&nw);
      wr= &a_wr;
    }

    // Indirection vectors containing the original indices
    std::vector<size_t, Malloc_allocator<size_t> >
      d((Malloc_allocator<size_t>(key_memory_JSON)));

    std::vector<size_t, Malloc_allocator<size_t> >
      c((Malloc_allocator<size_t>(key_memory_JSON)));

    // Sort both vectors, so we can compare efficiently
    sort_array(doc_wrapper, &d);
    sort_array(*wr, &c);

    size_t doc_i= 0;

    for (size_t c_i= 0; c_i < c.size() && doc_i < d.size(); c_i++)
    {
      Json_dom::enum_json_type candt= (*wr)[c[c_i]].type();

      if (candt == Json_dom::J_ARRAY)
      {
        while (doc_i < d.size() &&
               doc_wrapper[d[doc_i]].type() < candt)
        {
          doc_i++;
        }

        bool found= false;
        /*
          We do not increase doc_i here, use a tmp. We might need to check again
          against doc_i: this allows duplicates in the candidate.
        */
        for (size_t tmp= doc_i;
             tmp < d.size() && doc_wrapper[d[tmp]].type() == Json_dom::J_ARRAY;
             tmp++)
        {
          if (contains_wr(doc_wrapper[d[tmp]], (*wr)[c[c_i]]))
          {
            found= true;
            break;
          }
        }

        if (!found)
          return false;
      }
      else
      {
        bool found= false;
        size_t tmp= doc_i;

        while (tmp < d.size())
        {
          if (doc_wrapper[d[tmp]].type() == Json_dom::J_ARRAY ||
              doc_wrapper[d[tmp]].type() == Json_dom::J_OBJECT)
          {
            if (contains_wr(doc_wrapper[d[tmp]], (*wr)[c[c_i]]))
            {
              found= true;
              break;
            }
          }
          else if (doc_wrapper[d[tmp]].compare((*wr)[c[c_i]]) == 0)
          {
            found= true;
            break;
          }
          tmp++;
        }

        if (doc_i == d.size() || !found)
            return false;
      }
    }

    return true;
  }
  else
  {
    return doc_wrapper.compare(containee_wr) == 0;
  }
}

longlong Item_func_json_contains::val_int()
{
  DBUG_ENTER("Item_func_json_contains::val_int");
  DBUG_ASSERT(fixed == 1);

  try {
    Json_wrapper doc_wrapper;

    // arg 0 is the document
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &doc_wrapper) ||
        args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(0);
    }

    Json_wrapper containee_wr;

    // arg 1 is the possible containee
    if (!get_json_wrapper(args, 1, &m_doc_value, func_name(), &containee_wr) ||
        args[1]->null_value)
    {
      null_value= true;
      DBUG_RETURN(0);
    }

    if (arg_count == 3)
    {
      // path is specified
      Json_path path;
      if (parse_path(args[2], &m_path_value, true, &path))
      {
        null_value= true;
        DBUG_RETURN(0);
      }

      Json_wrapper_vector
        v((Json_wrapper_vector::allocator_type(key_memory_JSON)));
      doc_wrapper.seek(path, &v, true);

      if (v.size() == 0)
      {
        null_value= true;
        DBUG_RETURN(1);
      }

      if (contains_wr(v[0], containee_wr))
      {
        null_value= false;
        DBUG_RETURN(1);
      }
    }
    else
    {
      if (contains_wr(doc_wrapper, containee_wr))
      {
        null_value= false;
        DBUG_RETURN(1);
      }
    }
  } CATCH_ALL("jsn_contains", DBUG_RETURN(error_int()))

  null_value= false;
  DBUG_RETURN(0);
}


longlong Item_func_json_contains_path::val_int()
{
  DBUG_ENTER("Item_func_json_contains_path::val_int");
  DBUG_ASSERT(fixed == 1);
  longlong result= 0;
  null_value= false;

  Json_wrapper wrapper;
  Json_wrapper_vector
    hits((Json_wrapper_vector::allocator_type(key_memory_JSON)));

  try {
    // arg 0 is the document
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &wrapper) ||
        args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(0);
    }

    // arg 1 is the oneOrAll flag
    String *const one_or_all= args[1]->val_str(&m_one_or_all_value);
    if (!one_or_all || args[1]->null_value)
    {
      null_value= true;
      DBUG_RETURN(0);
    }
    bool require_all;
    switch (parse_one_or_all(one_or_all->ptr(), one_or_all->length()))
    {
    case ooa_all:
      {
        require_all= true;
        break;
      }
    case ooa_one:
      {
        require_all= false;
        break;
      }
    default:
      {
        my_error(ER_JSON_BAD_CONTAINS_PATH_ARG, MYF(0));
        DBUG_RETURN(0);
      }
    }

    // the remaining args are paths
    for (uint32 i= 2; i < arg_count; ++i)
    {
      Json_path path;
      if (parse_path(args[i], &m_path_value, false, &path)) // empty_path
      {
        null_value= true;
        DBUG_RETURN(0);
      }

      hits.clear();
      wrapper.seek(path, &hits, true);
      if (hits.size() > 0)
      {
        result= 1;
        if (!require_all)
        {
          break;
        }
      }
      else
      {
        if (require_all)
        {
          result= 0;
          break;
        }
      }
    }

  } CATCH_ALL("jsn_contains_path", DBUG_RETURN(error_int()))

  DBUG_RETURN(result);
}


/**
  Get a JSON value from the item indicated by arg_idx.
  @param[in,out] args    the arguments to the function
  @param[in]     arg_idx the argument index
  @param[out]    result  the JSON result value area

  @return true if a JSON value (or NULL) is found, else false
*/
bool json_value(Item **args, uint arg_idx, Json_wrapper *result)
{
  Item *arg= args[arg_idx];
  bool ok= true;

  if (arg->field_type() == MYSQL_TYPE_NULL)
  {
    DBUG_ASSERT(arg->null_value);
  }
  else if (arg->field_type() != MYSQL_TYPE_JSON)
  {
    // This is not a JSON value. Give up.
    ok= false;
  }
  else
  {
    ok= arg->val_json(result);
  }

  return ok;
}


/**
  Get a JSON value in args[args_idx], either by an existing JSON
  value, or by parsing a string as a JSON text.

  @param[in]     args        the arguments
  @param[in]     arg_idx     the argument index
  @param[in,out] str         string buffer to use
  @param[in]     func_name   the name of the function we are executing
  @param[in,out] wr          the resulting JSON value location

  @result true if we found a value or NULL, false if not.
*/
bool get_json_wrapper(Item **args,
                      uint arg_idx,
                      String *str,
                      const char *func_name,
                      Json_wrapper *wrapper)
{

  DBUG_ENTER("get_json_wrapper");

  if (json_value(args, arg_idx, wrapper))
  {
    // Found a JSON value, return successfully.
    DBUG_RETURN(true);
  }

  if (args[arg_idx]->field_type() == MYSQL_TYPE_JSON)
  {
    /*
      If the type of the argument is JSON and json_value() returned
      false, it means the argument didn't contain valid JSON data.
      Give up.
    */
    DBUG_RETURN(false);
  }

  /*
    Otherwise, it's a non-JSON type, so we need to see if we can
    convert it to JSON.
  */

  if (args[arg_idx]->null_value)
  {
    DBUG_RETURN(true);
  }

  /* Is this a JSON text? */
  bool needed_parse= false ; //@< true indicates a JSON string
  Json_dom *dom; //@< we'll receive a DOM here from a successful text parse

  if (!json_is_valid(args, arg_idx, str, func_name, &needed_parse, &dom, true))
  {
    DBUG_RETURN(false);
  }

  if (args[arg_idx]->null_value)
  {
    DBUG_RETURN(true);
  }

  DBUG_ASSERT(needed_parse);
  DBUG_ASSERT(dom);

  *wrapper= Json_wrapper(dom);
  DBUG_RETURN(true);
}

/**
   Compute an index into json_type_string_map
   to be applied to certain sub-types of J_OPAQUE.

   @param field_type[in] The refined field type of the opaque value.

   @return an index into json_type_string_map
*/
static uint opaque_index(enum_field_types field_type)
{
  uint offset= 0;

  switch(field_type)
  {
  case MYSQL_TYPE_TINY_BLOB:
  case MYSQL_TYPE_MEDIUM_BLOB:
  case MYSQL_TYPE_LONG_BLOB:
  case MYSQL_TYPE_BLOB:
  case MYSQL_TYPE_STRING:
    {
      offset= static_cast<uint>(Json_dom::J_OPAQUE_BLOB);
      break;
    }

  case MYSQL_TYPE_BIT:
    {
      offset= static_cast<uint>(Json_dom::J_OPAQUE_BIT);
      break;
    }

  case MYSQL_TYPE_GEOMETRY:
    {
      offset= static_cast<uint>(Json_dom::J_OPAQUE_GEOMETRY);
      break;
    }

  default:
    {
      return static_cast<uint>(Json_dom::J_OPAQUE);
    }
  }

  return 1 + static_cast<uint>(Json_dom::J_ERROR) + offset;
}

String *Item_func_json_type::val_str(String *str)
{
  DBUG_ENTER("Item_func_json_type::val_str");
  DBUG_ASSERT(fixed == 1);

  try {
    Json_wrapper wr;
    if (!get_json_wrapper(args, 0, str, func_name(), &wr) ||
        args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(NULL);
    }

    const Json_dom::enum_json_type type= wr.type();
    uint typename_idx= static_cast<uint>(type);
    if (type == Json_dom::J_OPAQUE)
    {
      typename_idx= opaque_index(wr.field_type());
    }

    str->length(0);
    str->append(Json_dom::json_type_string_map[typename_idx]);

  } CATCH_ALL("jsn_type", str->length(0))

  null_value= false;
  DBUG_RETURN(str);
}


String *Item_json_func::val_str(String *str)
{
  DBUG_ENTER("Item_json_func::val_str");
  DBUG_ASSERT(fixed == 1);
  Json_wrapper wr;
  val_json(&wr);

  if (null_value)
    DBUG_RETURN(NULL);

  str->length(0);

  if (wr.type() != Json_dom::J_ERROR &&
      wr.to_string(str, true))
  {
    null_value= false;
    DBUG_RETURN(str);
  }

  my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
  DBUG_RETURN(error_str());
}


bool Item_json_func::get_date(MYSQL_TIME *ltime, my_time_flags_t fuzzydate)
{
  Json_wrapper wr;
  if (!val_json(&wr))
    return true;
  
  if (null_value)
    return true;
  
  switch (wr.type())
  {
  case Json_dom::J_DATE:
  case Json_dom::J_TIMESTAMP:
  case Json_dom::J_DATETIME:
  case Json_dom::J_TIME:
    {
      set_zero_time(ltime, MYSQL_TIMESTAMP_DATETIME);
      wr.get_datetime(ltime);

      if (ltime->time_type == MYSQL_TIMESTAMP_TIME)
      {
        MYSQL_TIME tmp= *ltime;
        time_to_datetime(current_thd, &tmp, ltime);
      }

      null_value= false;
      return false;
    }
  default:;
  }
  
  null_value= true;
  return true; // error
}


bool Item_json_func::get_time(MYSQL_TIME *ltime)
{
  Json_wrapper wr;
  if (!val_json(&wr))
    return true;
  
  if (null_value)
    return true;
  
  switch (wr.type())
  {
    case Json_dom::J_DATETIME:
    case Json_dom::J_DATE:
    case Json_dom::J_TIME:
    case Json_dom::J_TIMESTAMP:
    {
      set_zero_time(ltime, MYSQL_TIMESTAMP_DATETIME);
      wr.get_datetime(ltime);
      null_value= false;
      return false;
    }
    default:;
  }
  

  char offending_value_buff[MAX_FIELD_WIDTH];
  String ov_str(offending_value_buff, sizeof(offending_value_buff),
                &my_charset_utf8mb4_bin);
  ov_str.length(0);
  wr.to_string(&ov_str, true);

  if (ov_str.length() >= MAX_FIELD_WIDTH)
    ov_str.length(MAX_FIELD_WIDTH - 1); // print no more

  push_warning_printf(current_thd,
                      Sql_condition::SL_WARNING,
                      ER_TRUNCATED_WRONG_VALUE,
                      ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                      "DATE/TIME/DATETIME/TIMESTAMP",
                      ov_str.c_ptr_safe(),
                      func_name(),
                      0);
  null_value= true;
  return true; // error
}


longlong Item_json_func::val_int()
{
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_utf8mb4_bin);
  Json_wrapper wr;
  if (!val_json(&wr))
    return 0;
  
  if (null_value)
    return 0;
  
  switch (wr.type())
  {
  case Json_dom::J_UINT:
    {
      ulonglong ui= wr.get_uint();
      if (ui > static_cast<ulonglong>(std::numeric_limits<longlong>::max()))
      {
        push_warning_printf(current_thd, Sql_condition::SL_ERROR,
                            ER_WARN_DATA_OUT_OF_RANGE,
                            ER_THD(current_thd, ER_WARN_DATA_OUT_OF_RANGE),
                            item_name.ptr(), 1L);
      }
      return static_cast<longlong>(ui);
    }
  case Json_dom::J_INT:
    {
      longlong i= wr.get_int();
      return i;
    }
  case Json_dom::J_STRING:
    {
      /*
       For a string result, we must first get the string and then convert it
       to a longlong
       */
      char buff[MAX_FIELD_WIDTH], *end;
      String res(buff, sizeof(buff), &my_charset_utf8mb4_bin);
      if (res.copy(wr.get_data(), wr.get_data_length(),
                   &my_charset_utf8mb4_bin))
      {
        my_error(ER_OUT_OF_RESOURCES, MYF(0));
        null_value= true;
        return 0;
      }

      null_value= false;

      char *start= (char *)res.ptr();
      size_t length= res.length();
      const CHARSET_INFO *cs= res.charset();

      end= start + length;
      int error;
      longlong value= cs->cset->strtoll10(cs, start, &end, &error);

      if (error > 0 || end != start+ length)
      {
        ErrConvString err(&res);
        push_warning_printf(current_thd, Sql_condition::SL_WARNING,
                            ER_TRUNCATED_WRONG_VALUE,
                            ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                            "INTEGER",
                            err.ptr());
      }
      return value;
    }
  case Json_dom::J_BOOLEAN:
    {
      return wr.get_boolean() ? 1 : 0;
    }
  case Json_dom::J_DECIMAL:
    {
      longlong i;
      my_decimal decimal_value;
      wr.get_decimal_data(&decimal_value);
      /*
       We do not know if this int is destined for signed or unsigned usage, so
       just get longlong from the value using the sign in the decimal.
       */
      my_decimal2int(E_DEC_FATAL_ERROR, &decimal_value,
                     !decimal_value.sign(), &i);
      return i;
    }
  case Json_dom::J_DOUBLE:
    {
      // logic here is borrowed from Field_double::val_int
      double j= wr.get_double();
      longlong res;

      if (j <= (double) LLONG_MIN)
      {
        res= (longlong) LLONG_MIN;
      }
      else if (j >= (double) (ulonglong) LLONG_MAX)
      {
        res= (longlong) LLONG_MAX;
      }
      else
      {
        return (longlong) rint(j);
      }

      char buffer[FLOATING_POINT_BUFFER];
      String s(buffer, sizeof(buffer), &my_charset_utf8mb4_bin);
      s.length(0);
      wr.to_string(&s, true);
      ErrConvString err(&s);
      push_warning_printf(current_thd,
                          Sql_condition::SL_WARNING,
                          ER_TRUNCATED_WRONG_VALUE,
                          ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                          "INTEGER",
                          err.ptr());
      return res;
    }
  default:;
  }

  char buffer[MAX_FIELD_WIDTH];
  String str2(buffer, sizeof(buffer), &my_charset_utf8mb4_bin);
  str2.length(0);
  wr.to_string(&str2, true);
  size_t to_print= std::min(static_cast<size_t>(str2.length()),
                            static_cast<size_t>(MAX_FIELD_WIDTH - 1));
  str2[to_print]= '\0'; // print no more

  push_warning_printf(current_thd,
                      Sql_condition::SL_WARNING,
                      ER_TRUNCATED_WRONG_VALUE,
                      ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                      "INT",
                      str2.ptr());
  null_value= true;
  return 0;
}


double Item_json_func::val_real()
{
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_utf8mb4_bin);
  Json_wrapper wr;
  if (!val_json(&wr))
    return 0.0;

  if (null_value)
    return 0.0;

  switch (wr.type())
  {
    case Json_dom::J_DECIMAL:
    {
      double dbl;
      my_decimal decimal_value;
      wr.get_decimal_data(&decimal_value);
      my_decimal2double(E_DEC_FATAL_ERROR, &decimal_value, &dbl);
      return dbl;
    }
  case Json_dom::J_STRING:
    {
      /*
       For a string result, we must first get the string and then convert it
       to a double
       */
      char buff[MAX_FIELD_WIDTH], *end;
      String res(buff, sizeof(buff), &my_charset_utf8mb4_bin);
      if (res.copy(wr.get_data(), wr.get_data_length(),
                   &my_charset_utf8mb4_bin))
      {
        my_error(ER_OUT_OF_RESOURCES, MYF(0));
        return 0;
      }

      char *start= (char *)res.ptr();
      size_t length= res.length();
      const CHARSET_INFO *cs= res.charset();

      end= start + length;
      int error;
      double value= cs->cset->strntod(cs, start, length, &end, &error);

      if (error || end != start + length)
      {
        ErrConvString err(&res);
        if (error)
        {
          push_warning_printf(current_thd, Sql_condition::SL_WARNING,
                              ER_TRUNCATED_WRONG_VALUE,
                              ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                              "DOUBLE",
                              err.ptr());
        }
        else
        {
          push_warning_printf(current_thd, Sql_condition::SL_WARNING,
                              WARN_DATA_TRUNCATED,
                              ER_THD(current_thd, WARN_DATA_TRUNCATED),
                              func_name(),
                              current_thd->get_stmt_da()->
                                current_row_for_condition());
        }
     }
      return value;
    }
    case Json_dom::J_DOUBLE:
      return wr.get_double();
    case Json_dom::J_INT:
      return static_cast<double>(wr.get_int());
    case Json_dom::J_UINT:
      return static_cast<double>(wr.get_uint());
    case Json_dom::J_BOOLEAN:
      return static_cast<double>(wr.get_boolean());
    default:;
  }

  char buffer[MAX_FIELD_WIDTH];
  String str2(buffer, sizeof(buffer), &my_charset_utf8mb4_bin);
  str2.length(0);
  wr.to_string(&str2, true);
  size_t to_print= std::min(static_cast<size_t>(str2.length()),
                            static_cast<size_t>(MAX_FIELD_WIDTH - 1));
  str2[to_print]= '\0'; // print no more

  push_warning_printf(current_thd,
                      Sql_condition::SL_WARNING,
                      ER_TRUNCATED_WRONG_VALUE,
                      ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                      "DOUBLE",
                      str2.ptr());

  null_value= true;
  return 0.0;
}

my_decimal *Item_json_func::val_decimal(my_decimal *decimal_value)
{
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_utf8mb4_bin);
  Json_wrapper wr;

  if (!val_json(&wr))
    return decimal_value;

  if (null_value)
    return decimal_value;

  switch (wr.type())
  {
    case Json_dom::J_DECIMAL:
    {
      wr.get_decimal_data(decimal_value);
      return decimal_value;
    }
    case Json_dom::J_STRING:
    {
      /*
       For a string result, we must first get the string and then convert it
       to a decimal
       */
      char buff[MAX_FIELD_WIDTH];
      String res(buff, sizeof(buff), &my_charset_utf8mb4_bin);

      if (res.copy(wr.get_data(), wr.get_data_length(),
                   &my_charset_utf8mb4_bin))
      {
        my_error(ER_OUT_OF_RESOURCES, MYF(0));
        return 0;
      }

      int err= str2my_decimal(E_DEC_FATAL_ERROR, res.ptr(),
                              res.length(), res.charset(), decimal_value);
      if (err)
      {
        ErrConvString errmsg(res.ptr(), res.length(), res.charset());
        push_warning_printf(current_thd, Sql_condition::SL_WARNING,
                            ER_TRUNCATED_WRONG_VALUE,
                            ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                            "DECIMAL", errmsg.ptr());
      }
      return decimal_value;
    }

    case Json_dom::J_DOUBLE:
    {
      double d= wr.get_double();
      double2my_decimal(E_DEC_FATAL_ERROR, d, decimal_value);
      return decimal_value;
    }
    case Json_dom::J_INT:
    {
      longlong l= wr.get_int();
      longlong2decimal(l, decimal_value);
      return decimal_value;
    }
    case Json_dom::J_UINT:
    {
      longlong l= wr.get_uint();
      longlong2decimal(l, decimal_value);
      return decimal_value;
    }
    case Json_dom::J_BOOLEAN:
    {
      longlong i= wr.get_boolean();
      int2my_decimal(E_DEC_FATAL_ERROR, i, true /* 0 or 1 */, decimal_value);
      return decimal_value;
    }
    default:;
  }

  char offending_value_buff[MAX_FIELD_WIDTH];
  String ov_str(offending_value_buff, sizeof(offending_value_buff),
                &my_charset_utf8mb4_bin);
  ov_str.length(0);
  wr.to_string(&ov_str, true);

  if (ov_str.length() >= MAX_FIELD_WIDTH)
    ov_str.length(MAX_FIELD_WIDTH - 1); // print no more

  push_warning_printf(current_thd,
                      Sql_condition::SL_WARNING,
                      ER_TRUNCATED_WRONG_VALUE,
                      ER_THD(current_thd, ER_TRUNCATED_WRONG_VALUE),
                      "DECIMAL",
                      ov_str.c_ptr_safe(),
                      func_name(),
                      0);
  my_decimal_set_zero(decimal_value);
  return decimal_value;

}


/**
  Get a JSON value from a function, field or subselect scalar.

  @param[in]     arg         the function argument
  @param[in,out] value       a scratch area
  @param[in,out] tmp         temporary scratch space for converting strings to
                             the correct charset; only used if accept_string is
                             true and conversion is needed
  @param[out]    wr          the retrieved JSON value
  @param[in,out] scalar      pointer to pre-allocated memory that can be
                             borrowed by the result wrapper to hold the scalar
                             result. If the pointer is NULL, memory will be
                             allocated on the heap.
  @param[in]     accept_string
                             if true, accept SQL strings as scalars
                             (false implies we need a valid
                              JSON parsable string)
  @return true if we could get a value or NULL, otherwise false
*/
bool val_json_func_field_subselect(Item* arg,
                                   String *value,
                                   String *tmp,
                                   Json_wrapper *wr,
                                   Json_scalar_holder *scalar,
                                   bool accept_string)
{
  enum_field_types  field_type= arg->field_type();
  Json_dom *dom= NULL;

  /*
    TINYTEXT, TEXT, MEDIUMTEXT, and LONGTEXT have type
    MYSQL_TYPE_BLOB. We want to treat them like strings.
    We check the collation to see if the blob is really
    a string.
  */
  if (((field_type == MYSQL_TYPE_TINY_BLOB) ||
       (field_type == MYSQL_TYPE_BLOB) ||
       (field_type == MYSQL_TYPE_MEDIUM_BLOB) ||
       (field_type == MYSQL_TYPE_LONG_BLOB))
      && (arg->collation.collation != &my_charset_bin))
  {
    field_type= MYSQL_TYPE_STRING;
  }

  switch (field_type)
  {
  case MYSQL_TYPE_INT24:
  case MYSQL_TYPE_LONG:
  case MYSQL_TYPE_SHORT:
  case MYSQL_TYPE_TINY:
  case MYSQL_TYPE_LONGLONG:
    {
      longlong i= arg->val_int();

      if (arg->null_value)
        return true;

      if (arg->unsigned_flag)
      {
        if (scalar)
        {
          scalar->m_val= Json_uint(i);
        }
        else
        {
          dom= new Json_uint(i);
        }
      }
      else if (scalar)
      {
        scalar->m_val= Json_int(i);
      }
      else
      {
        dom= new Json_int(i);
      }

      break;
    }
  case MYSQL_TYPE_DATE:
  case MYSQL_TYPE_DATETIME:
  case MYSQL_TYPE_TIMESTAMP:
  case MYSQL_TYPE_TIME:
    {
      longlong dt= arg->val_temporal_by_field_type();

      if (arg->null_value)
        return true;

      MYSQL_TIME t;
      TIME_from_longlong_datetime_packed(&t, dt);
      t.time_type= field_type_to_timestamp_type(field_type);
      if (scalar)
      {
        scalar->m_val= Json_datetime(t, field_type);
      }
      else
      {
        dom= new Json_datetime(t, field_type);
      }
      break;
    }
  case MYSQL_TYPE_NEWDECIMAL:
    {
      my_decimal m;
      my_decimal *r= arg->val_decimal(&m);

      if (arg->null_value)
        return true;

      if (!r)
      {
        my_error(ER_INVALID_CAST_TO_JSON, MYF(0));
        return false;
      }

      if (scalar)
      {
        scalar->m_val= Json_decimal(*r);
      }
      else
      {
        dom= new Json_decimal(*r);
      }
      break;
    }
  case MYSQL_TYPE_DOUBLE:
  case MYSQL_TYPE_FLOAT:
    {
      double d= arg->val_real();

      if (arg->null_value)
        return true;

      if (scalar)
      {
        scalar->m_val= Json_double(d);
      }
      else
      {
        dom= new Json_double(d);
      }
      break;
    }
  case MYSQL_TYPE_BLOB:
  case MYSQL_TYPE_BIT:
  case MYSQL_TYPE_GEOMETRY:
  case MYSQL_TYPE_LONG_BLOB:
  case MYSQL_TYPE_MEDIUM_BLOB:
  case MYSQL_TYPE_TINY_BLOB:
  case MYSQL_TYPE_YEAR:
    {
      String *oo= arg->val_str(value);

      if (arg->null_value)
        return true;

      enum_field_types ft= arg->field_type();

      if (arg->type() == Item::FIELD_ITEM && ft == MYSQL_TYPE_BLOB)
      {
        Item_field *it_f= down_cast<Item_field *>(arg);
        Field *f= it_f->field;
        Field_blob *fb= down_cast<Field_blob *>(f);
        switch (fb->pack_length() - portable_sizeof_char_ptr)
        {
        case 1:
          ft= MYSQL_TYPE_TINY_BLOB;
          break;
        case 2:
          ft= MYSQL_TYPE_BLOB;
          break;
        case 3:
          ft= MYSQL_TYPE_MEDIUM_BLOB;
          break;
        case 4:
          ft= MYSQL_TYPE_LONG_BLOB;
          break;
        default:
          DBUG_ABORT();
        }
      }

      if (scalar)
      {
        scalar->m_val= Json_opaque(ft, oo->ptr(), oo->length());
      }
      else
      {
        dom= new Json_opaque(ft, oo->ptr(), oo->length());
      }
      break;
    }
  case MYSQL_TYPE_VAR_STRING:
  case MYSQL_TYPE_VARCHAR:
  case MYSQL_TYPE_ENUM:
  case MYSQL_TYPE_SET:
  case MYSQL_TYPE_STRING:
    {
      /*
        Wrong charset or Json syntax error (the latter: only if !accept_string,
        in which case a binary character set is our only hope for success).
      */
      String *res= arg->val_str(value);

      if (arg->null_value)
        return true;
      const CHARSET_INFO *cs= res->charset();

      if (cs == &my_charset_bin)
      {
        // BINARY or similar
        enum_field_types ft= (arg->type() == Item::VARBIN_ITEM) ?
          MYSQL_TYPE_BIT : arg->field_type();
        if (scalar)
        {
          scalar->m_val= Json_opaque(ft, res->ptr(), res->length());
        }
        else
        {
          dom= new Json_opaque(ft, res->ptr(), res->length());
        }
        break;
      }
      else if (accept_string)
      {
        const char *s= res->ptr();
        size_t ss= res->length();

        if (!ensure_utf8mb4(res, tmp, &s, &ss, true))
        {
          return false;
        }

        if (scalar)
        {
          scalar->m_val= Json_string(std::string(s, ss));
        }
        else
        {
          dom= new Json_string(std::string(s, ss));
        }
      }
      else
      {
        my_error(ER_INVALID_CAST_TO_JSON, MYF(0));
        return false;
      }
      break;
    }
  case MYSQL_TYPE_DECIMAL:                      // pre 5.0
    my_error(ER_NOT_SUPPORTED_YET, MYF(0), "old decimal type");
    return false;

  case MYSQL_TYPE_NULL:                         // possible?
    DBUG_ASSERT(arg->null_value);
    return true;

  case MYSQL_TYPE_JSON:
    DBUG_ABORT();                               /* purecov: inspected */
    // fall-through
  default:
    my_error(ER_INVALID_CAST_TO_JSON, MYF(0));
    return false;
  }

  // Exactly one of scalar and dom should be used.
  DBUG_ASSERT((scalar == NULL) != (dom == NULL));
  DBUG_ASSERT((scalar == NULL) || (scalar->get() != NULL));

  Json_wrapper w(scalar ? scalar->get() : dom);
  if (scalar)
  {
    /*
      The DOM object lives in memory owned by the caller. Tell the
      wrapper that it's not the owner.
    */
    w.set_alias();
  }
  wr->steal(&w);

  return true;
}


/**
  Try to determine whether an argument has a boolean (as opposed
  to an int) type, and if so, return its boolean value.

  @param[in] arg The argument to inspect.
  @param[in/out] result Fill in the result if this is a boolean arg.

  @return True if the arg can be determined to have a boolean type.
*/
bool extract_boolean(Item *arg, bool *result)
{
  if (arg->is_bool_func())
  {
    *result= arg->val_int();
    return true;
  }
  else if (arg->type() == Item::SUBSELECT_ITEM)
  {
    // EXISTS subqueries have boolean type
    Item_exists_subselect *ies= dynamic_cast<Item_exists_subselect *>(arg);
    if (ies != NULL)
    {
      *result= arg->val_int();
      return true;
    }
  }
  else if (arg->type() == Item::FUNC_ITEM)
  {
    // gtid_subset() returns a boolean
    Item_func_gtid_subset *func= dynamic_cast<Item_func_gtid_subset *>(arg);
    if (func != NULL)
    {
      *result= arg->val_int();
      return true;
    }
  }

  if (arg->type() == Item::INT_ITEM)
  {
    const Name_string * const name= &arg->item_name;
    const bool is_literal_false= name->is_set() && name->eq("FALSE");
    const bool is_literal_true= name->is_set() && name->eq("TRUE");
    if (is_literal_false || is_literal_true)
    {
      *result= is_literal_true;
      return true;
    }
  }

  // doesn't fit any of the checks we perform
  return false;
}


bool get_json_atom_wrapper(Item **args,
                           uint arg_idx,
                           String *value,
                           String *tmp,
                           Json_wrapper *wr,
                           Json_scalar_holder *scalar,
                           bool accept_string)
{
  bool result= true;

  Item * const arg= args[arg_idx];

  try
  {
    if (json_value(args, arg_idx, wr))
    {
      return true;
    }

    if (arg->field_type() == MYSQL_TYPE_JSON)
    {
      /*
        If the type of the argument is JSON and json_value() returned
        false, it means the argument didn't contain valid JSON data.
        Give up.
      */
      return false;
    }

    // boolean operators should produce boolean values
    bool  boolean_value;
    if (extract_boolean(arg, &boolean_value))
    {
      Json_dom *boolean_dom;
      if (scalar)
      {
        scalar->m_val= Json_boolean(boolean_value);
        boolean_dom= scalar->get();
      }
      else
      {
        boolean_dom= new Json_boolean(boolean_value);
      }
      Json_wrapper wrapper(boolean_dom);
      if (scalar)
      {
        /*
          The DOM object lives in memory owned by the caller. Tell the
          wrapper that it's not the owner.
        */
        wrapper.set_alias();
      }
      wr->steal(&wrapper);
      return true;
    }

    /*
      Allow other types as first-class or opaque JSON values.
      But how to determine what the type is? We do a best effort...
    */
    result= val_json_func_field_subselect(arg, value, tmp, wr, scalar,
                                          accept_string);

  } CATCH_ALL("CAST(... AS JSON)", result= false)

  return result;
}


bool Item_json_typecast::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_typecast::val_json");
  DBUG_ASSERT(fixed == 1);

  bool needed_parse= false ; //@< true indicates a JSON string, not a binary
  Json_dom *dom= NULL;       //@< if non-null we want a DOM from parse

  if (args[0]->field_type() == MYSQL_TYPE_NULL)
  {
    null_value= true;
    DBUG_RETURN(true);
  }
  else if (args[0]->field_type() == MYSQL_TYPE_JSON)
  {
    if (!json_value(args, 0, wr))
      DBUG_RETURN(error_json());

    if (args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }
    null_value= false;
    DBUG_RETURN(true);
  }
  else if  (json_is_valid(args, 0, &m_value, func_name(),
                          &needed_parse, &dom, false))
  {
    if (args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }
    // We were able to parse a JSON value from a string.
    DBUG_ASSERT(needed_parse);
    DBUG_ASSERT(dom);
    // Pass on the DOM wrapped
    Json_wrapper w(dom);
    wr->steal(&w);
    null_value= false;
    DBUG_RETURN(true);
  }

  // Not a string, nor a JSON value, wrap the rest
  args[0]->null_value= false; // a priori: new attempt

  if (!get_json_atom_wrapper(args, 0, &m_value, &m_conversion_buffer,
                             wr, NULL, true))
    DBUG_RETURN(error_json());

  null_value= args[0]->null_value;
  DBUG_RETURN(true);
}


void Item_json_typecast::print(String *str, enum_query_type query_type)
{
  str->append(STRING_WITH_LEN("cast("));
  args[0]->print(str, query_type);
  str->append(STRING_WITH_LEN(" as "));
  str->append(cast_type());
  str->append(')');
}


longlong Item_func_json_length::val_int()
{
  DBUG_ENTER("Item_func_json_length::val_int");
  DBUG_ASSERT(fixed == 1);
  longlong result= 0;

  Json_wrapper wrapper;

  try {
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &wrapper) ||
        args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(0);
    }
  } CATCH_ALL("jsn_length", DBUG_RETURN(0))

  if (arg_count > 1)
  {
    Json_path json_path;
    if (parse_path(args[1], &m_path_value, true, &json_path))
    {
      null_value= true;
      DBUG_RETURN(0);
    }

    Json_wrapper_vector
      hits((Json_wrapper_vector::allocator_type(key_memory_JSON)));
    wrapper.seek(json_path, &hits, true);

    if (hits.size() != 1)
    {
      // path does not exist. return null.
      null_value= true;
      DBUG_RETURN(0);
    }

    // there should only be one hit because wildcards were forbidden
    DBUG_ASSERT(hits.size() == 1);

    wrapper.steal(&hits[0]);
  }

  result= wrapper.length();

  null_value= false;
  DBUG_RETURN(result);
}


longlong Item_func_json_depth::val_int()
{
  DBUG_ENTER("Item_func_json_depth::val_int");
  DBUG_ASSERT(fixed == 1);
  longlong result= 0;

  Json_wrapper wrapper;

  try {
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &wrapper) ||
        args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(0);
    }
  } CATCH_ALL("jsn_depth", DBUG_RETURN(error_int()))

  result= wrapper.depth();

  null_value= false;
  DBUG_RETURN(result);
}


bool Item_func_json_keys::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_keys::val_json");
  DBUG_ASSERT(fixed == 1);

  Json_wrapper wrapper;

  try {
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &wrapper))
      DBUG_RETURN(error_json());
    if (args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }

    if (arg_count > 1)
    {
      Json_path path;
      if (parse_path(args[1], &m_path_value, true, &path))
      {
        null_value= true;
        DBUG_RETURN(true);
      }
      Json_wrapper_vector
        hits((Json_wrapper_vector::allocator_type(key_memory_JSON)));
      wrapper.seek(path, &hits, false);

      if (hits.size() != 1)
      {
        null_value= true;
        DBUG_RETURN(true);
      }

      wrapper.steal(&hits[0]);
    }

    if (wrapper.type() != Json_dom::J_OBJECT)
    {
      null_value= true;
      DBUG_RETURN(true);
    }

    // We have located a JSON object value, now collect its keys
    // and return them as a JSON array.
    Json_array *res= new Json_array();
    for (Json_wrapper_object_iterator i(wrapper.object_iterator());
         !i.empty(); i.next())
    {
      Json_string s(i.elt().first);
      res->append_clone(&s);
    }
    Json_wrapper resw(res);
    wr->steal(&resw);

  } CATCH_ALL("jsn_keys", DBUG_RETURN(error_json()))

  null_value= false;
  DBUG_RETURN(true);
}


bool Item_func_json_extract::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_extract::val_json");
  DBUG_ASSERT(fixed == 1);

  try
  {
    Json_wrapper w;
    // collect results here
    Json_wrapper_vector
      v((Json_wrapper_vector::allocator_type(key_memory_JSON)));

    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &w))
      DBUG_RETURN(error_json());

    if (args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }

    for (uint32 i= 1; i < arg_count; ++i)
    {
      Json_path path;
      if (parse_path(args[i], &m_path_value, false, &path)) // empty path
      {
        null_value= true;
        DBUG_RETURN(true);
      }

      w.seek(path, &v, true);
    }

    if (v.size() == 0)
    {
      null_value= true;
      DBUG_RETURN(true);
    }
    else if (v.size() == 1)
    {
      wr->steal(&v[0]);
    }
    else
    {
      Json_array *a= new Json_array();
      for (Json_wrapper_vector::iterator it= v.begin(); it != v.end(); ++it)
      {
        a->append_clone(it->to_dom());
      }
      Json_wrapper w(a);
      wr->steal(&w);
    }

  } CATCH_ALL("jsn_extract", DBUG_RETURN(error_json()))

  null_value= false;
  DBUG_RETURN(true);
}


bool Item_func_json_append::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_append::val_json");
  DBUG_ASSERT(fixed == 1);

  if (arg_count % 2 == 0) // 3,5,7, ..., (k*2)+1 args allowed
  {
    my_error(ER_WRONG_PARAMCOUNT_TO_NATIVE_FCT, MYF(0), func_name());
    DBUG_RETURN(error_json());
  }

  try {
    Json_wrapper docw;

    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &docw))
      DBUG_RETURN(error_json());
    if (args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }

    for (uint32 i= 1; i < arg_count; i += 2)
    {
      // Need a DOM to be able to manipulate arrays
      Json_dom *doc= docw.to_dom();

      Json_path path;
      if (parse_path(args[i], &m_path_value, true, &path))
      {
        // empty path (error signalled already)
        null_value= true;
        DBUG_RETURN(true);
      }

      Json_dom_vector hits((Json_dom_vector::allocator_type(key_memory_JSON)));
      doc->seek(path, &hits, true);

      if (hits.size() < 1)
      {
        continue;
      }

      /*
        Use a reverse iterator lest we get into trouble with replacing outer
        parts of the doc before we get to paths to inner parts when we have
        ellipses in the path. Make sure we do the most nested replacements
        first. Json_dom::seek returns outermost hits first.
      */
      for (Json_dom_vector::reverse_iterator it= hits.rbegin();
           it != hits.rend(); ++it)
      {
        Json_wrapper valuew;
        if (!get_json_atom_wrapper(args, i + 1, &m_value, &m_conversion_buffer,
                                   &valuew, NULL, true))
          DBUG_RETURN(error_json());
        if (args[i + 1]->null_value)
        {
          null_value= true;
          DBUG_RETURN(true);
        }

        if ((*it)->json_type() == Json_dom::J_ARRAY)
        {
          Json_array *arr= down_cast<Json_array *>(*it);
          arr->append_alias(valuew.to_dom());
          valuew.set_alias(); // we have taken over the DOM
        }
        else
        {
          Json_array *arr= new Json_array();
          arr->append_clone(*it);
          arr->append_alias(valuew.to_dom());
          valuew.set_alias(); // we have taken over the DOM
          /*
            This value will replace the old document we found using path, since
            we did an auto-wrap. If this is root, this is trivial, but if it's
            inside an array or object, we need to find the parent DOM to be
            able to replace it in situ.
          */
          if (path.leg_count() == 0) // root
          {
            Json_wrapper newroot(arr);
            docw.steal(&newroot);
          } else
          {
            Json_dom *parent= (*it)->parent();
            DBUG_ASSERT(parent);
            parent->replace_dom_in_container(*it, arr);
          }
        }
      }
    }

    // docw still owns the augmented doc, so hand it over to result
    wr->steal(&docw);

  } CATCH_ALL("jsn_append", DBUG_RETURN(error_json()))

  null_value= false;
  DBUG_RETURN(true);
}


bool Item_func_json_insert::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_insert::val_json");
  DBUG_ASSERT(fixed == 1);

  if (arg_count % 2 == 0) // 3,5,7, ..., (k*2)+1 args allowed
  {
    my_error(ER_WRONG_PARAMCOUNT_TO_NATIVE_FCT, MYF(0), func_name());
    DBUG_RETURN(error_json());
  }

  try {
    Json_wrapper docw;

    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &docw))
      DBUG_RETURN(error_json());

    if(args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }

    for (uint32 i= 1; i < arg_count; i += 2)
    {
      // Need a DOM to be able to manipulate arrays and objects
      Json_dom *doc= docw.to_dom();

      Json_path path;
      if (parse_path(args[i], &m_path_value, true, &path))
      {
        // empty path (error signalled already)
        null_value= true;
        DBUG_RETURN(true);
      }

      {
        Json_dom_vector
          hits((Json_dom_vector::allocator_type(key_memory_JSON)));
        doc->seek(path, &hits, false);

        if (hits.size() != 0 || // already exists
            path.leg_count() == 0) // is root
        {
          continue;
        }
      }

      /*
        Need to look one step up the path: if we are specifying an array slot
        we need to find the array. If we are specifying an object element, we
        need to find the object. In both cases so we can insert into them.

        Remove the first path leg and search again.
      */
      Json_dom_vector hits((Json_dom_vector::allocator_type(key_memory_JSON)));
      Json_path_leg l= path.pop();
      doc->seek(path, &hits, false);

      if (hits.size() < 1)
      {
        // no unique object found at parent position, so bail out
        continue;
      }

      Json_wrapper valuew;
      if (!get_json_atom_wrapper(args, i + 1, &m_value, &m_conversion_buffer,
                                 &valuew, NULL, true))
      {
        DBUG_RETURN(error_json());
      }

      if(args[i + 1]->null_value)
      {
        null_value= true;
        DBUG_RETURN(true);
      }

      /*
        Use a reverse iterator lest we get into trouble with replacing outer
        parts of the doc before we get to paths to inner parts when we have
        ellipses in the path. Make sure we do the most nested replacements
        first. Json_dom::seek returns outermost hits first.
      */
      for (Json_dom_vector::reverse_iterator it= hits.rbegin();
           it != hits.rend(); ++it)
      {
        // We found *something* at that parent path

        // What did we specify in the path, object or array?
        if (l.get_type() == jpl_array_cell)
        {
          // We specified an array, what did we find at that position?
          if ((*it)->json_type() == Json_dom::J_ARRAY)
          {
            Json_array *arr= down_cast<Json_array *>(*it);
            DBUG_ASSERT(l.get_type() == jpl_array_cell);
            arr->insert_clone(l.get_array_cell_index(), valuew.to_dom());
          }
          else if (l.get_array_cell_index() > 0)
          {
            /*
              Found a scalar or object and we didn't specify position 0:
              auto-wrap it
            */
            Json_dom *a= *it;
            Json_array *newarr= new Json_array();
            newarr->append_clone(a); //auto-wrap this
            newarr->insert_clone(l.get_array_cell_index(), valuew.to_dom());

            /*
              Now we need this value to replace the old document we found using
              path. If this is root, this is trivial, but if it's inside an
              array or object, we need to find the parent DOM to be able to
              replace it in situ.
            */
            if (path.leg_count() == 0) // root
            {
              Json_wrapper newroot(newarr);
              docw.steal(&newroot);
            } else
            {
              Json_dom *parent= a->parent();
              DBUG_ASSERT(parent);

              parent->replace_dom_in_container(a, newarr);
            }
          }
        }
        else if (l.get_type() == jpl_member &&
                 (*it)->json_type() == Json_dom::J_OBJECT)
        {
          Json_object *o= down_cast<Json_object *>(*it);
          const char *ename= l.get_member_name();
          size_t enames= l.get_member_name_length();
          o->add_clone(std::string(ename, enames), valuew.to_dom());
        }
      }
    }
    // docw still owns the augmented doc, so hand it over to result
    wr->steal(&docw);

  } CATCH_ALL("jsn_insert", DBUG_RETURN(error_json()))

  null_value= false;
  DBUG_RETURN(true);
}

/**
  Common implementation for JSN_SET and JSN_REPLACE
*/
bool Item_func_json_set_replace::val_json(Json_wrapper *wr)
{

  if (arg_count % 2 == 0) // 3,5,7, ..., (k*2)+1 args allowed
  {
    my_error(ER_WRONG_PARAMCOUNT_TO_NATIVE_FCT, MYF(0), func_name());
    return error_json();
  }

  try {
    Json_wrapper docw;

    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &docw))
      return error_json();

    if(args[0]->null_value)
    {
      null_value= true;
      return true;
    }

    Json_dom_set hit_set((Json_dom_set::key_compare()),
                         Json_dom_set::allocator_type(key_memory_JSON));

    for (uint32 i= 1; i < arg_count; i += 2)
    {
      // Need a DOM to be able to manipulate arrays and objects
      Json_dom *doc= docw.to_dom();

      Json_path path;
      if (parse_path(args[i], &m_path_value, true /* allow wildcards */,
                     &path))
      {
        // empty path (error signalled already)
        null_value= true;
        return true;
      }

      Json_dom_vector hits((Json_dom_vector::allocator_type(key_memory_JSON)));
      doc->seek(path, &hits, false);


      Json_wrapper valuew;
      if (!get_json_atom_wrapper(args, i + 1, &m_value, &m_conversion_buffer,
                                 &valuew, NULL, true))
        return error_json();

      if(args[i + 1]->null_value)
      {
        null_value= true;
        return true;
      }

      if (hits.size() == 0)
      {
        /*
          Need to look one step up the path: if we are specifying an array slot
          we need to find the array. If we are specifying an object element, we
          need to find the object. In both cases so we can insert into them.

          Remove the first path leg and search again.
        */
        Json_path_leg l= path.pop();
        doc->seek(path, &hits, false);

        if (hits.size() < 1)
        {
          // no unique object found at parent position, so bail out
          continue;
        }

        /*
          Use a reverse iterator lest we get into trouble with replacing outer
          parts of the doc before we get to paths to inner parts when we have
          ellipses in the path. Make sure we do the most nested replacements
          first. Json_dom::seek returns outermost hits first.
        */
        for (Json_dom_vector::reverse_iterator it= hits.rbegin();
            it != hits.rend(); ++it)
        {
          // We now have either an array or an object in the parent's path
          if (l.get_type() == jpl_array_cell)
          {
            if ((*it)->json_type() == Json_dom::J_ARRAY)
            {
              if (!m_jsn_set) // replace semantics, so skip if path not present
                continue;

              Json_array *arr= down_cast<Json_array *>(*it);
              DBUG_ASSERT(l.get_type() == jpl_array_cell);
              arr->insert_clone(l.get_array_cell_index(), valuew.to_dom());
            }
            else
            {
              /*
                Found a scalar or object, auto-wrap it and make it the first
                element in a new array, unless the new value specifies position
                0, in which case the old gets replaced.
              */
              Json_dom *a= *it;
              Json_dom *res;

              if (l.get_array_cell_index() == 0)
              {
                res= valuew.to_dom()->clone();
              }
              else
              {
                // replace semantics, so we don't get larger array
                if (!m_jsn_set)
                  continue;

                Json_array *newarr= new Json_array();
                newarr->append_clone(a);
                newarr->insert_clone(l.get_array_cell_index(), valuew.to_dom());
                res= newarr;
              }

              /*
                Now we need this value to replace the old document we found
                using path. If this is root, this is trivial, but if it's
                inside an array or object, we need to find the parent DOM to be
                able to replace it in situ.
              */
              if (path.leg_count() == 0) // root
              {
                Json_wrapper newroot(res);
                docw.steal(&newroot);
              } else
              {
                Json_dom *parent= a->parent();
                DBUG_ASSERT(parent);

                parent->replace_dom_in_container(a, res);
              }
            }
          }
          else if (l.get_type() == jpl_member &&
                   (*it)->json_type() == Json_dom::J_OBJECT)
          {
            if (!m_jsn_set) // replace semantics, so skip if path not present
              continue;

            Json_object *o= down_cast<Json_object *>(*it);
            const char *ename= l.get_member_name();
            size_t enames= l.get_member_name_length();
            o->add_clone(std::string(ename, enames), valuew.to_dom());
          }
        }
      }
      else
      {
        // we found one or more value, so replace semantics.
        hit_set.clear();
        /*
          The seek() method may identify many matches, some of which
          are nested inside other matches. Consider the following situation:

          jsn_remove( '[ { "a": { "a": true } } ]', '$[**].a' )

          So we throw away a match if any of its ancestors are also matches
          since they'd be replaced anyway.
        */

        // make a priori set for lookup
        for (Json_dom_vector::iterator it= hits.begin(); it != hits.end(); ++it)
        {
          hit_set.insert(*it);
        }

        /*
          Run through the hits again and check each one if it has a parent
          in the a priori set. If so, delete it from the hits.
        */
        for (Json_dom_vector::iterator it= hits.begin(); it != hits.end(); ++it)
        {
          for (Json_dom *candidate= (*it)->parent();
               candidate != NULL; candidate= candidate->parent())
          {
            if (hit_set.find(candidate) != hit_set.end())
            {
              *it= NULL;
              break;
            }
          }
        }

        // remaining hits are valid paths
        for (Json_dom_vector::iterator it= hits.begin(); it != hits.end(); ++it)
        {
          Json_dom *child= *it;

          // skip matches which were pruned out above
          if (child == NULL)
            continue;

          Json_dom *parent= child->parent();
          if (!parent)
          {
            Json_wrapper w(valuew.to_dom()->clone());
            docw.steal(&w);
          }
          else
          {
            parent->replace_dom_in_container(child, valuew.to_dom()->clone());
          }
        }
      } // if: found 1 else more values
    } // do: functions argument list run-though

    // docw still owns the augmented doc, so hand it over to result
    wr->steal(&docw);

  } CATCH_ALL(func_name(), return error_json())

  null_value= false;
  return true;
}


bool Item_func_json_array::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_array::val_json");
  DBUG_ASSERT(fixed == 1);

  try {
    Json_array *arr= new Json_array();
    Json_wrapper docw(arr);

    for (uint32 i= 0; i < arg_count; ++i)
    {
      Json_wrapper valuew;
      if (!get_json_atom_wrapper(args, i, &m_value, &m_conversion_buffer,
                                 &valuew, NULL, true))
      {
        DBUG_RETURN(error_json());
      }

      if (args[i]->null_value)
      {
        null_value= true;
        DBUG_RETURN(true);
      }

      arr->append_alias(valuew.to_dom());
      valuew.set_alias(); // release the DOM
    }

    // docw still owns the augmented doc, so hand it over to result
    wr->steal(&docw);

  } CATCH_ALL("jsn_array", DBUG_RETURN(error_json()))

  null_value= false;
  DBUG_RETURN(true);
}

bool Item_func_json_row_object::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_row_object::val_json");
  DBUG_ASSERT(fixed == 1);

  try {
    Json_object *object= new Json_object();
    Json_wrapper docw(object);

    for (uint32 i= 0; i < arg_count; ++i)
    {
      /*
        arguments come in pairs. we have already verified that there
        are an even number of args.
      */
      uint32  key_idx= i++;
      uint32  value_idx= i;

      // key
      Item *key_item= args[key_idx];
      std::string key;
      char buff[MAX_FIELD_WIDTH];
      String utf8_res(buff, sizeof(buff), &my_charset_utf8mb4_bin);
      const char *safep;         // contents of key_item, possibly converted
      size_t safe_length;        // length of safep

      if (!get_json_string(key_item, &tmp_key_value, &utf8_res, func_name(),
                           &safep, &safe_length))
      {
        null_value= true;
        DBUG_RETURN(true);
      }
      key.append( safep, safe_length );

      // value
      Json_wrapper valuew;
      if (!get_json_atom_wrapper(args, value_idx, &m_value,
                                 &m_conversion_buffer, &valuew, NULL, true))
      {
        DBUG_RETURN(error_json());
      }

      if ((args[key_idx]->null_value) || (args[value_idx]->null_value))
      {
        null_value= true;
        DBUG_RETURN(true);
      }

      object->add_alias(key, valuew.to_dom());
      valuew.set_alias(); // release the DOM
    }

    // docw still owns the augmented doc, so hand it over to result
    wr->steal(&docw);

  } CATCH_ALL("jsn_object", DBUG_RETURN(error_json()))

  null_value= false;
  DBUG_RETURN(true);
}

bool Item_func_json_search::satisfies_like(const char *string_data,
                                           size_t length)
{
  m_source_string->set_str_with_copy(string_data, static_cast<uint>(length));

  if (m_like_node->val_int() == 1)
  {
    return true;
  }

  return false;
}

void Item_func_json_search::find_matches(const Json_wrapper &wrapper,
                                         Json_path *path,
                                         Json_dom_vector *matches,
                                         String_set *duplicates,
                                         bool one_match)
{
  switch (wrapper.type())
  {
  case Json_dom::J_STRING:
    {
      if (one_match && !matches->empty())
      {
        return;
      }

      if (satisfies_like(wrapper.get_data(), wrapper.get_data_length()))
      {
        std::string string_contents= path->to_string();
        std::pair<std::set<std::string>::iterator, bool> res=
          duplicates->insert(string_contents);

        if (res.second)
        {
          matches->push_back(new Json_string(string_contents));
        }
      }
      break;
    }

  case Json_dom::J_OBJECT:
    {
      for (Json_wrapper_object_iterator jwot(wrapper.object_iterator());
           !jwot.empty(); jwot.next())
      {
        std::pair<const std::string, Json_wrapper> pair= jwot.elt();
        const std::string key= pair.first;
        Json_wrapper value= pair.second;
        Json_path_leg next_leg(key);

        // recurse
        path->append(next_leg);
        find_matches(value, path, matches, duplicates, one_match);
        path->pop();

        if (one_match && !matches->empty())
        {
          return;
        }
      }
      break;
    }

  case Json_dom::J_ARRAY:
    {
      for (size_t idx= 0; idx < wrapper.length(); idx++)
      {
        Json_wrapper value= wrapper[idx];
        Json_path_leg next_leg(idx);

        // recurse
        path->append(next_leg);
        find_matches(value, path, matches, duplicates, one_match);
        path->pop();

        if (one_match && !matches->empty())
        {
          return;
        }
      }
      break;
    }

  default:
    {
      break;
    }
  } // end switch on wrapper type

  return;
}

bool Item_func_json_search::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_search::val_json");
  DBUG_ASSERT(fixed == 1);

  Json_dom_vector matches((Json_dom_vector::allocator_type(key_memory_JSON)));
  String_set duplicates((String_set::key_compare()),
                        String_set::allocator_type(key_memory_JSON));

  try {
    Json_wrapper docw;

    // arg 0 is the document
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &docw) ||
        args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(false);
    }

    // arg 1 is the oneOrAll arg
    String *const one_or_all= args[1]->val_str(&m_one_or_all_value);
    if (!one_or_all || args[1]->null_value)
    {
      null_value= true;
      DBUG_RETURN(false);
    }
    bool one_match;
    switch (parse_one_or_all(one_or_all->ptr(), one_or_all->length()))
    {
    case ooa_all:
      {
        one_match= false;
        break;
      }
    case ooa_one:
      {
        one_match= true;
        break;
      }
    default:
      {
        my_error(ER_JSON_BAD_SEARCH_ARG, MYF(0));
        DBUG_RETURN(false);
      }
    }

    // arg 2 is the LIKE node. set by the constructor

    // the remaining arguments are path expressions
    if (arg_count == 3) // no user-supplied path expressions
    {
        Json_path path;

        find_matches(docw, &path, &matches, &duplicates, one_match);
    }
    else  // user-supplied path expressions
    {
      Json_wrapper_vector
        hits((Json_wrapper_vector::allocator_type(key_memory_JSON)));
      Json_path_vector
        path_vector((Json_path_vector::allocator_type(key_memory_JSON)));

      // validate the user-supplied path expressions
      for (uint32 i= 3; i < arg_count; ++i)
      {
        Json_path path;
        if (parse_path(args[i], &m_path_value, false, &path)) // empty path
        {
          null_value= true;
          DBUG_RETURN(true);
        }

        path_vector.push_back(path);
      }

      // find the matches for each of the user-supplied path expressions
      for (Json_path_vector::iterator jpt= path_vector.begin();
           jpt != path_vector.end(); ++jpt)
      {
        if (one_match && (matches.size() > 0))
        {
          break;
        }

        Json_path path= *jpt;

        /*
          If there are wildcards in the path, then we need to
          compute the full path to the subdocument. We can only
          do this on doms.
        */
        if (path.contains_wildcard_or_ellipsis())
        {
          Json_dom *dom= docw.to_dom();
          Json_dom_vector
            dom_hits((Json_dom_vector::allocator_type(key_memory_JSON)));

          dom->seek(path, &dom_hits, false);

          for (Json_dom_vector::iterator jdvi= dom_hits.begin();
               jdvi != dom_hits.end(); ++jdvi)
          {
            if (one_match && (matches.size() > 0))
            {
              break;
            }

            Json_dom *subdocument= *jdvi;
            Json_path subdocument_path= subdocument->get_location();
            Json_wrapper subdocument_wrapper(subdocument);
            subdocument_wrapper.set_alias();

            find_matches(subdocument_wrapper, &subdocument_path,
                         &matches, &duplicates, one_match);
          } // end of loop through hits
        }
        else // no wildcards in the path
        {
          hits.clear();
          docw.seek(path, &hits, false);

          for (Json_wrapper_vector::iterator jwvi= hits.begin();
               jwvi != hits.end(); ++jwvi)
          {
            if (one_match && (matches.size() > 0))
            {
              break;
            }

            Json_wrapper  subdocument_wrapper= *jwvi;

            find_matches(subdocument_wrapper, &path, &matches, &duplicates,
                         one_match);
          } // end of loop through hits
        }  // end if the user-supplied path expression has wildcards
      }   // end of loop through user-supplied path expressions
    }     // end if there are user-supplied path expressions

  } CATCH_ALL("jsn_append", DBUG_RETURN(error_json()))

  if (matches.size() == 0)
  {
    null_value= true;
    DBUG_RETURN(true);
  }
  else if (matches.size() == 1)
  {
    Json_wrapper scalar_wrapper(matches[0]);
    wr->steal(&scalar_wrapper);
  }
  else
  {
    Json_array  *array= new Json_array();
    for (Json_dom_vector::iterator vsi= matches.begin();
         vsi != matches.end(); ++vsi)
    {
      array->append_alias(*vsi);
    }

    Json_wrapper  array_wrapper(array);
    wr->steal(&array_wrapper);
  }

  null_value= false;
  DBUG_RETURN(true);
}


bool Item_func_json_remove::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_remove::val_json");
  DBUG_ASSERT(fixed == 1);

  Json_wrapper wrapper;
  uint32  path_count= arg_count - 1;
  bool  had_error= false;
  null_value= false;

  try {
    if (!get_json_wrapper(args, 0, &m_doc_value, func_name(), &wrapper))
      DBUG_RETURN(error_json());
    if (args[0]->null_value)
    {
      null_value= true;
      DBUG_RETURN(true);
    }

    m_paths.clear();
    for (uint32 path_idx= 0; path_idx < path_count; ++path_idx)
    {
      m_paths.push_back(Json_path());
      if (parse_path(args[path_idx + 1], &m_path_value, true, &m_paths.back()))
      {
        null_value= true;
        break;
      }
    }

  } CATCH_ALL("jsn_remove", (had_error= true))

  if (had_error || null_value)
  {
    DBUG_RETURN(had_error ? error_json() : true);
  }

  for (Json_path_vector::iterator it= m_paths.begin();
       it != m_paths.end(); ++it)
  {
    if (it->leg_count() == 0)
    {
      my_error(ER_JSON_VACUOUS_PATH, MYF(0));
      DBUG_RETURN(error_json());
    }
  }

  // good document, good paths. do some work

  // no binary support for removal. must convert to a dom.
  Json_dom *dom= wrapper.to_dom();

  // remove elements identified by the paths, one after the other
  Json_dom_vector hits((Json_dom_vector::allocator_type(key_memory_JSON)));
  Json_dom_set hit_set((Json_dom_set::key_compare()),
                       Json_dom_set::allocator_type(key_memory_JSON));
  for (Json_path_vector::iterator path= m_paths.begin();
       path != m_paths.end(); ++path)
  {
    /*
      The seek() method may identify many matches, some of which
      are nested inside other matches. Consider the following situation:

      jsn_remove( '[ { "a": { "a": true } } ]', '$[**].a' )

      So we throw away a match if any of its ancestors are also matches.
    */
    hits.clear();
    hit_set.clear();

    // now find the matches
    dom->seek(*path, &hits, true);

    // remove matches whose ancestors are also matches
    for (Json_dom_vector::iterator it= hits.begin(); it != hits.end(); ++it)
    {
      hit_set.insert(*it);
    }

    for (Json_dom_vector::iterator it= hits.begin(); it != hits.end(); ++it)
    {
      // eliminate this child if one of its ancestors is also a match
      for (Json_dom *candidate= (*it)->parent();
           candidate != NULL; candidate= candidate->parent())
      {
        if (hit_set.find(candidate) != hit_set.end())
        {
          *it= NULL;
          break;
        }
      }
    }

    // now remove matches
    for (Json_dom_vector::iterator it= hits.begin(); it != hits.end(); ++it)
    {
      Json_dom *child= *it;

      // skip matches which were pruned out above
      if (child == NULL)
        continue;

      Json_dom *parent= child->parent();

      // no parent means the root. the path is nonsense.
      if (parent == NULL )
      {
        continue;
      }

      Json_dom::enum_json_type type= parent->json_type();
      DBUG_ASSERT((type == Json_dom::J_OBJECT) || (type == Json_dom::J_ARRAY));

      if (type == Json_dom::J_OBJECT)
      {
        Json_object *object= down_cast<Json_object *>(parent);
        object->remove(child);
      }
      else if (type == Json_dom::J_ARRAY)
      {
        Json_array *array= down_cast<Json_array *>(parent);
        array->remove(child);
      }
    } // end of loop through matches on current path
  }   // end of loop through all paths

  // wrapper still owns the pruned doc, so hand it over to result
  wr->steal(&wrapper);

  DBUG_RETURN(true);
}


bool Item_func_json_merge::val_json(Json_wrapper *wr)
{
  DBUG_ENTER("Item_json_func_merge::val_json");
  DBUG_ASSERT(fixed == 1);

  Json_dom *result_dom= NULL;
  bool had_error= false;
  null_value= false;

  try {
    for (uint idx= 0; idx < arg_count; idx++)
    {
      Json_wrapper next_wrapper;
      if (!get_json_wrapper(args, idx, &m_value, func_name(), &next_wrapper))
      {
        had_error= true;
        break;
      }

      if (args[idx]->null_value)
      {
        null_value= true;
        break;
      }

      /*
        Grab the next DOM, release it from its wrapper, and merge it
        into the previous DOM.
      */
      Json_dom *next_dom= next_wrapper.to_dom();
      next_wrapper.set_alias();
      result_dom= (idx == 0) ? next_dom : merge_doms(result_dom, next_dom);
    }
  } CATCH_ALL("jsn_merge", (had_error= true))

  if (had_error || null_value)
  {
    delete result_dom;
    DBUG_RETURN(had_error ? error_json() : true);
  }

  // fake a wrapper so that we can hand its dom to the return arg
  Json_wrapper tmp(result_dom);
  wr->steal(&tmp);
  DBUG_RETURN(true);
}


String *Item_func_json_quote::val_str(String *str)
{
  DBUG_ENTER("Item_json_func_quote::val_str");
  DBUG_ASSERT(fixed == 1);

  String *res= args[0]->val_str(str);

  try {
    const char *safep;
    size_t safep_size;

    if (!res)
    {
      null_value= true;
      DBUG_RETURN(NULL);
    }

    switch (args[0]->field_type())
    {
    case MYSQL_TYPE_STRING:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_VARCHAR:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_TINY_BLOB:
      break;
    default:
      my_error(ER_INCORRECT_TYPE, MYF(0), "1", func_name());
      DBUG_RETURN(error_str());
    }

    if (!ensure_utf8mb4(res, &m_value, &safep, &safep_size, true))
    {
      null_value= true;
      DBUG_RETURN(NULL);
    }

    std::string qt= double_quote(safep, safep_size);
    res->length(0);
    res->append(qt.c_str(), qt.size());
    res->set_charset(&my_charset_utf8mb4_bin);
  } CATCH_ALL("jsn_quote", DBUG_RETURN(error_str()))

  null_value= false;
  DBUG_RETURN(res);
}


String *Item_func_json_unquote::val_str(String *str)
{
  DBUG_ENTER("Item_func_json_unquote::val_str");
  DBUG_ASSERT(fixed == 1);

  Json_dom *dom= NULL;

  try {
    bool need_parse;

    if (args[0]->field_type() == MYSQL_TYPE_JSON)
    {
      Json_wrapper wr;
      if (!get_json_wrapper(args, 0, str, func_name(), &wr))
      {
        DBUG_RETURN(error_str());
      }

      if (args[0]->null_value)
      {
        null_value= true;
        DBUG_RETURN(NULL);
      }

      m_value.length(0);

      if (!wr.to_string(&m_value, false))
      {
        DBUG_RETURN(error_str());
      }

      null_value= false;
      DBUG_RETURN(&m_value);
    }


    String *res= args[0]->val_str(str);

    if (!res)
    {
      null_value= true;
      DBUG_RETURN(NULL);
    }

    /*
      We only allow a string argument, so get rid of any other
      type arguments.
    */
    switch (args[0]->field_type())
    {
    case MYSQL_TYPE_STRING:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_VARCHAR:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_TINY_BLOB:
      break;
    default:
      my_error(ER_INCORRECT_TYPE, MYF(0), "1", func_name());
      DBUG_RETURN(error_str());
    }

    if (res->length() < 2 || *res->ptr() != '"' ||
        res->ptr()[res->length() - 1] != '"')
    {
      null_value= false;
      DBUG_RETURN(res); // return string unchanged
    }

    if (!parse_json(res, args, 0, func_name(), &need_parse, &dom, true))
    {
      null_value= true;
      DBUG_RETURN(error_str());
    }

    /*
      Extract the internal string representation as a MySQL string
    */
    DBUG_ASSERT(need_parse);
    DBUG_ASSERT(dom->json_type() == Json_dom::J_STRING);
    std::string s= down_cast<Json_string *>(dom)->value();
    str->length(0);
    str->append(s.c_str(), s.size());
    str->set_charset(&my_charset_utf8mb4_bin);
    delete dom;
  } CATCH_ALL("jsn_unquote", delete dom; DBUG_RETURN(error_str()))


  null_value= false;
  DBUG_RETURN(str);
}


Json_scalar *Json_scalar_holder::get()
{
  return boost::polymorphic_get<Json_scalar>(&m_val);
}
