#ifndef ITEM_JSON_FUNC_INCLUDED
#define ITEM_JSON_FUNC_INCLUDED

/* Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */

#include "json_path.h"
#include "json_dom.h"
#include "item_strfunc.h"       // Item_str_func
#include "malloc_allocator.h"   // Malloc_allocator

typedef std::vector<Json_path, Malloc_allocator<Json_path> > Json_path_vector;
typedef std::set<std::string, std::less<std::string>,
                 Malloc_allocator<std::string> > String_set;

class Item_func_like;

/* JSON function support  */

/**
  Base class for all item functions that a return JSON value
*/
class Item_json_func : public Item_func
{
protected:
  /// String used when reading JSON binary values or JSON text values.
  String m_value;
  /// String used for converting JSON text values to utf8mb4 charset.
  String m_conversion_buffer;
public:
  Item_json_func(const POS &pos, Item *a) : Item_func(pos, a)
  {}
  Item_json_func(const POS &pos, Item *a, Item *b) : Item_func(pos, a, b)
  {}
  Item_json_func(const POS &pos, PT_item_list *a) : Item_func(pos, a)
  {}

  enum_field_types field_type() const { return MYSQL_TYPE_JSON; }
  type_conversion_status save_in_field(Field *field, bool no_conversions);

  void fix_length_and_dec()
  {
    max_length= MAX_BLOB_WIDTH;
    maybe_null= true;
    collation.collation= &my_charset_utf8mb4_bin;
  }
  enum Item_result result_type () const { return STRING_RESULT; }
  String *val_str(String *arg);
  bool get_date(MYSQL_TIME *ltime, my_time_flags_t fuzzydate);
  bool get_time(MYSQL_TIME *ltime);
  longlong val_int();
  double val_real();
  my_decimal *val_decimal(my_decimal *decimal_value);

  Item_result cast_to_int_type () const { return INT_RESULT; }

  void update_null_value ()
  {
    Json_wrapper wr;
    val_json(&wr);
  }
};

/**
  Return the JSON value of the argument in a wrapper. Abstracts whether
  the value comes from a field or a function. Does not handle literals.
  See also get_json_wrapper.

  @param[in]     args     the arguments
  @param[in]     arg_idx  the argument index
  @param[in,out] result   the JSON value wrapper

  @returns true iff the argument is a JSON field or function result
*/
extern bool json_value(Item **args, uint arg_idx, Json_wrapper *result);

/**
  Return the JSON value of the argument in a wrapper. Abstracts whether
  the value comes from a field or a function or a valid JSON text.

  @param[in]  args          the arguments
  @param[in]  arg_idx       the argument index
  @param[out] str           the string buffer
  @param[in]  func_name     the name of the function we are executing
  @param[out] result        the JSON value wrapper

  @returns true iff the argument is a JSON value or NULL. If it is not
           NULL, the value is available in result (the wrapper).
           Returns false the value isn't JSON. No error message is
           generated; it is the responsibility of the caller.
*/
extern bool get_json_wrapper(Item **args,
                             uint arg_idx,
                             String *str,
                             const char *func_name,
                             Json_wrapper *wrapper);

/**
  Check a non-empty val for character set. If it has character set
  my_charset_binary, signal error and return false. Else, try to convert to
  my_charset_utf8mb4_binary. If this fails, signal error and return false, else
  return true.

  @param[in]     val       the string to be checked
  @param[in,out] buf       buffer to hold the converted string
  @param[out]    resptr    the resulting, possibly converted string,
                           only set if no error
  @param[out]    reslength the length of resptr
  @param[in]     require_string
                           If true, give error messages if binary string. If we
                           see a conversion error (space), we give error
                           notwithstanding this parameter value
  @return true if no error, else false
*/
bool ensure_utf8mb4(String *val,
                    String *buf,
                    const char **resptr,
                    size_t *reslength,
                    bool require_string);

/**
  Represents the JSON function JSN_VALID( <value> )
*/
class Item_func_json_valid :public Item_int_func
{
  String m_value;
public:
  Item_func_json_valid(const POS &pos, Item *a) : Item_int_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_valid";
  }

  bool is_bool_func()
  {
    return 1;
  }

  longlong val_int();

  void fix_length_and_dec()
  {
    maybe_null= true;
  }
};

/**
  Represents the JSON function JSN_CONTAINS()
*/
class Item_func_json_contains :public Item_int_func
{
  String m_doc_value;
  String m_one_or_all_value;
  String m_path_value;
public:
  Item_func_json_contains(const POS &pos, PT_item_list *a)
    : Item_int_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_contains";
  }

  bool is_bool_func()
  {
    return 1;
  }

  longlong val_int();

  void fix_length_and_dec()
  {
    maybe_null= true;
  }
};

/**
  Represents the JSON function JSN_CONTAINS_PATH()
*/
class Item_func_json_contains_path :public Item_int_func
{
  String m_doc_value;
  String m_one_or_all_value;
  String m_path_value;
public:
  Item_func_json_contains_path(const POS &pos, PT_item_list *a)
    : Item_int_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_contains_path";
  }

  bool is_bool_func()
  {
    return 1;
  }

  longlong val_int();

  void fix_length_and_dec()
  {
    maybe_null= true;
  }
};

/**
  Represents the JSON function JSN_TYPE
*/
class Item_func_json_type :public Item_str_func
{
  String m_value;
public:
  Item_func_json_type(const POS &pos, Item *a) : Item_str_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_type";
  }

  void fix_length_and_dec()
  {
    maybe_null= true;
    max_length= Json_dom::typelit_max_length;
  };

  String *val_str(String *);
};

/**
  Represents a "CAST( <value> AS JSON )" coercion.
*/
class Item_json_typecast :public Item_json_func
{
public:
  Item_json_typecast(const POS &pos, Item *a) : Item_json_func(pos, a)
  {}

  void print(String *str, enum_query_type query_type);
  const char *func_name() const { return "cast_as_json"; }
  const char *cast_type() const { return "json"; }
  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_LENGTH()
*/
class Item_func_json_length :public Item_int_func
{
  String m_doc_value;
  String m_path_value;

public:
  Item_func_json_length(const POS &pos, Item *a)
    : Item_int_func(pos, a)
  {}

  Item_func_json_length(const POS &pos, Item *a, Item *b)
    : Item_int_func(pos, a, b)
  {}

  void fix_length_and_dec()
  {
    maybe_null= true;
  }

  const char *func_name() const
  {
    return "jsn_length";
  }

  longlong val_int();
};

/**
  Represents the JSON function JSN_DEPTH()
*/
class Item_func_json_depth :public Item_int_func
{
  String m_doc_value;
  String m_path_value;

public:
  Item_func_json_depth(const POS &pos, Item *a)
    : Item_int_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_depth";
  }

  longlong val_int();
};

/**
  Represents the JSON function JSN_KEYS()
*/
class Item_func_json_keys :public Item_json_func
{
  String m_doc_value;
  String m_path_value;

public:
  Item_func_json_keys(const POS &pos, Item *a)
    : Item_json_func(pos, a)
  {}

  Item_func_json_keys(const POS &pos, Item *a, Item *b)
    : Item_json_func(pos, a, b)
  {}

  const char *func_name() const
  {
    return "jsn_keys";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_EXTRACT()
*/
class Item_func_json_extract :public Item_json_func
{
  String m_doc_value;
  String m_path_value;

public:
  Item_func_json_extract(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_extract";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_APPEND()
*/
class Item_func_json_append :public Item_json_func
{
  String m_doc_value;
  String m_path_value;
public:
  Item_func_json_append(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_append";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_INSERT()
*/
class Item_func_json_insert :public Item_json_func
{
  String m_doc_value;
  String m_path_value;
public:
  Item_func_json_insert(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_insert";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Common base class for JSN_SET() and JSN_REPLACE().
*/
class Item_func_json_set_replace :public Item_json_func
{
  /// True if this is JSN_SET, false if it is JSN_REPLACE.
  const bool m_jsn_set;
  String m_doc_value;
  String m_path_value;
protected:
  Item_func_json_set_replace(const POS &pos, PT_item_list *a, bool jsn_set)
    : Item_json_func(pos, a), m_jsn_set(jsn_set)
  {}
public:
  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_SET()
*/
class Item_func_json_set :public Item_func_json_set_replace
{
public:
  Item_func_json_set(const POS &pos, PT_item_list *a)
    : Item_func_json_set_replace(pos, a, true)
  {}

  const char *func_name() const
  {
    return "jsn_set";
  }
};

/**
  Represents the JSON function JSN_REPLACE()
*/
class Item_func_json_replace :public Item_func_json_set_replace
{
public:
  Item_func_json_replace( const POS &pos, PT_item_list *a )
    : Item_func_json_set_replace(pos, a, false)
  {}

  const char *func_name() const
  {
    return "jsn_replace";
  }
};

/**
  Represents the JSON function JSN_ARRAY()
*/
class Item_func_json_array :public Item_json_func
{
public:
  Item_func_json_array(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_array";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_OBJECT()
*/
class Item_func_json_row_object :public Item_json_func
{
  String tmp_key_value;
public:
  Item_func_json_row_object(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_object";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_SEARCH()
*/
class Item_func_json_search :public Item_json_func
{
  String m_doc_value;
  String m_one_or_all_value;
  String m_search_string_value;
  String m_path_value;

  // LIKE machinery
  Item_string *m_source_string;
  Item_func_like *m_like_node;
public:
  /**
   Construct a JSN_SEARCH() node.

   @param[in] pos Parser position
   @param[in] a Nodes which must be fixed (i.e. bound/resolved)
   @param[in] source Node to hold the source string being evaluted
   @param[in] like LIKE node which operates on source

   @returns a JSN_SEARCH() node.
  */
  Item_func_json_search(const POS &pos, PT_item_list *a,
                        Item_string *source, Item_func_like *like)
    : Item_json_func(pos, a), m_source_string(source), m_like_node(like)
  {}


  const char *func_name() const
  {
    return "jsn_search";
  }

  bool val_json(Json_wrapper *wr);

private:

  /**
   Evaluate whether a candidate string satisfies a LIKE clause.

   @param[in] string_data Character data of string to evaluate.
   @param[in] length Length of that character data.
 */
  bool satisfies_like(const char *string_data, size_t length);

  /**
   Recursive function to find the string values, nested inside
   a json document, which satisfy the LIKE condition. As matches
   are found, their path locations are added to an evolving
   vector of matches.

   @param[in] subdoc A subdocument of the original document.
   @param[in] path The path location of the subdocument
   @param[in,out] matches The evolving vector of matches.
   @param[in,out] duplicates List of paths found already.
   @param[in] one_match If true, then terminate search after first match.
 */
  void find_matches(const Json_wrapper &wrapper,
                    Json_path *path,
                    Json_dom_vector *matches,
                    String_set *duplicates,
                    bool one_match);
};

/**
  Represents the JSON function JSN_REMOVE()
*/
class Item_func_json_remove :public Item_json_func
{
  String m_doc_value;
  String m_path_value;
  Json_path_vector m_paths;

public:
  Item_func_json_remove(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a),
      m_paths(Json_path_vector::allocator_type(key_memory_JSON))
  {}

  const char *func_name() const
  {
    return "jsn_remove";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_MERGE()
*/
class Item_func_json_merge :public Item_json_func
{
public:
  Item_func_json_merge(const POS &pos, PT_item_list *a)
    : Item_json_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_merge";
  }

  bool val_json(Json_wrapper *wr);
};

/**
  Represents the JSON function JSN_QUOTE()
*/
class Item_func_json_quote :public Item_str_func
{
  String m_value;
public:
  Item_func_json_quote(const POS &pos, PT_item_list *a)
    : Item_str_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_quote";
  }

  void fix_length_and_dec()
  {
    maybe_null= true;
    max_length= args[0]->max_length;
  };

  String *val_str(String *tmpspace);
};

/**
  Represents the JSON function JSN_UNQUOTE()
*/
class Item_func_json_unquote :public Item_str_func
{
  String m_value;
public:
  Item_func_json_unquote(const POS &pos, PT_item_list *a)
    : Item_str_func(pos, a)
  {}

  const char *func_name() const
  {
    return "jsn_unquote";
  }

  void fix_length_and_dec()
  {
    maybe_null= true;
    max_length= args[0]->max_length;
  };

  String *val_str(String *str);
};

#endif /* ITEM_JSON_FUNC_INCLUDED */
