#ifndef JSON_ATOM_INCLUDED
#define JSON_ATOM_INCLUDED

/* Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */

/**
  @file json_atom.h

  This file provides the declarations of get_json_atom_wrapper() and
  Json_scalar_holder. These declarations have been moved into a
  separate header file because they depend on Boost, and we don't want
  this dependency to leak out to more files than strictly necessary.
*/

#include <json_dom.h>                           // Json_scalar
#include <sql_alloc.h>                          // Sql_alloc
#include <boost/variant.hpp>                    // boost::variant

/**
  Type that is capable of holding objects of any sub-type of
  Json_scalar. Used for pre-allocating space in query-duration memory
  for JSON scalars that are to be returned by get_json_atom_wrapper().

  Note: boost::blank is included in the variant to ensure that it
  includes a type that is known to be nothrow default-constructible.
  The presence of such a type avoids heap allocation when assigning a
  new value to the variant. Look for the "never-empty" guarantee in
  the Boost documentation for details.
*/
struct Json_scalar_holder : public Sql_alloc
{
  boost::variant<boost::blank, Json_string, Json_decimal, Json_int, Json_uint,
                 Json_double, Json_boolean, Json_null, Json_datetime,
                 Json_opaque> m_val;

  /**
    Get a pointer to the Json_scalar object contained in this holder.
    @return a pointer to a Json_scalar, or NULL if the holder is empty
  */
  Json_scalar *get();
};

/**
  Convert Json values or MySQL values to JSON.

  @param[in]     args       arguments to function
  @param[in]     arg_idx    the index of the argument to process
  @param[in,out] value      working area (if the returned Json_wrapper points
                            to a binary value rather than a DOM, this string
                            will end up holding the binary representation, and
                            it must stay alive until the wrapper is destroyed
                            or converted from binary to DOM)
  @param[in,out] tmp        temporary scratch space for converting strings to
                            the correct charset; only used if accept_string is
                            true and conversion is needed
  @param[in,out] wr         the result wrapper
  @param[in,out] scalar     pointer to pre-allocated memory that can be
                            borrowed by the result wrapper if the result is a
                            scalar. If the pointer is NULL, memory for a
                            scalar result will be allocated on the heap.
  @param[in]                accept_string
                            if true, accept MySQL strings as JSON strings
                            by converting them to UTF8, else emit an error
  @retval true    a JSON value or NULL was found
  @retval false   is error and needs signalling by the caller most likely
*/
bool get_json_atom_wrapper(Item **args, uint arg_idx, String *value,
                           String *tmp, Json_wrapper *wr,
                           Json_scalar_holder *scalar, bool accept_string);

#endif /* JSON_ATOM_INCLUDED */
