/* Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */

#include "json_dom.h"
#include "json_path.h"
#include "sql_time.h"
#include "base64.h"
#include "m_string.h"           // my_gcvt
#include "mysqld_error.h"       // ER_*
#include "template_utils.h"     // down_cast

#include "rapidjson/rapidjson.h"
#include "rapidjson/reader.h"

#include <sstream>              // std::stringbuf
#include <algorithm>            // std::min, std::max

using namespace rapidjson;

const char * Json_dom::json_type_string_map[]= {
  "NULL",
  "DECIMAL",
  "INTEGER",
  "UNSIGNED INTEGER",
  "DOUBLE",
  "STRING",
  "OBJECT",
  "ARRAY",
  "BOOLEAN",
  "DATE",
  "TIME",
  "DATETIME",
  "TIMESTAMP",
  "OPAQUE",
  "ERROR",

  // OPAQUE types with special names
  "BLOB",
  "BIT",
  "GEOMETRY",

  NULL
};

/**
  A dom is mergeable if it is an array or an object. All other
  types must be wrapped in an array in order to be merged.
*/
Json_dom *make_mergeable(Json_dom *candidate)
{
  switch (candidate->json_type())
  {
  case Json_dom::J_ARRAY:
  case Json_dom::J_OBJECT:
    {
      return candidate;
    }
  default:
    {
      return new Json_array(candidate);
    }
  }
}

Json_dom *merge_doms(Json_dom *left, Json_dom *right)
{
  left= make_mergeable(left);
  right= make_mergeable(right);

  // at this point, the arguments are either objects or arrays
  bool left_is_array= (left->json_type() == Json_dom::J_ARRAY);
  bool right_is_array= (right->json_type() == Json_dom::J_ARRAY);

  if (left_is_array || right_is_array)
  {
    if (!left_is_array)
      left= new Json_array(left);
    if (!right_is_array)
      right= new Json_array(right);

    down_cast<Json_array *>(left)->consume(down_cast<Json_array *>(right));
  }
  else  // otherwise, both doms are objects
  {
    down_cast<Json_object *>(left)->consume(down_cast<Json_object *>(right));
  }

  return left;
}


void *Json_dom::operator new(size_t size)
{
  return my_malloc(key_memory_JSON, size, MYF(MY_WME));
}


void Json_dom::operator delete(void *ptr)
{
  my_free(ptr);
}


/**
  Compute the maximum length of the string representation of the Json type
  literals which we use as output from JSN_TYPE.

  @return the length of the longest literal + 1 (for terminating NUL).
*/
static uint32 compute_max_typelit()
{
  size_t maxl= 0;
  for (const char **s= &Json_dom::json_type_string_map[0]; *s; ++s)
  {
    maxl= std::strlen(*s) > maxl ? std::strlen(*s) : maxl;
  }
  return static_cast<uint32>(maxl + 1);
}

const uint32 Json_dom::typelit_max_length= compute_max_typelit();

static bool seen_already(Json_dom_vector *result, Json_dom *cand)
{
  Json_dom_vector::iterator it= std::find(result->begin(),
                                          result->end(),
                                          cand);
  return it != result->end();
}

/**
 Add a value to a vector if it isn't already there.

 @param[in] candidate value to add
 @param[in/out] duplicates set of values added
 @param[in/out] result vector
 */
static void add_if_missing(Json_dom *candidate,
                           Json_dom_set *duplicates,
                           Json_dom_vector *result)
{
  if (duplicates->find(candidate) == duplicates->end())
  {
    result->push_back(candidate);
    duplicates->insert(candidate);
  }
}

void Json_dom::find_child_doms(const Json_path_leg *path_leg,
                               bool recursing,
                               bool auto_wrap,
                               Json_dom_set *duplicates,
                               Json_dom_vector *result)
{
  Json_dom::enum_json_type dom_type= json_type();
  enum_json_path_leg_type leg_type= path_leg->get_type();
  switch (leg_type)
  {
  case jpl_array_cell:
    {
      size_t array_cell_index= path_leg->get_array_cell_index();

      if (dom_type == Json_dom::J_ARRAY)
      {
        const Json_array * const array= down_cast<const Json_array *>(this);

        if (array_cell_index < array->size())
        {
          add_if_missing((*array)[array_cell_index], duplicates, result);
        }
      }
      else if (array_cell_index == 0 && auto_wrap)
      {
        if (!seen_already(result, this))
        {
          // auto-wrap non-arrays
          add_if_missing(this, duplicates, result);
        }
      }

      break;
    }
  case jpl_ellipsis:
    {
      add_if_missing(this, duplicates, result);

      if (dom_type == Json_dom::J_ARRAY)
      {
        const Json_array * const array= down_cast<const Json_array *>(this);

        for (unsigned eidx= 0; eidx < array->size(); eidx++)
        {
          Json_dom * child= (*array)[eidx];
          add_if_missing(child, duplicates, result);

          Json_dom::enum_json_type child_type= child->json_type();
          if ((child_type == Json_dom::J_ARRAY) ||
              (child_type == Json_dom::J_OBJECT))
          {
            // now recurse and add all objects and arrays under the child
            child->find_child_doms(path_leg, true, auto_wrap,
                                   duplicates, result);
          }
        } // end of loop through children
      }
      else if (dom_type == Json_dom::J_OBJECT)
      {
        const Json_object *const object=
          down_cast<const Json_object *>(this);

        for (Json_object_iterator iter= object->iterator();
             !iter.empty();
             iter.next())
        {
          Json_dom * child= iter.elt().second;
          Json_dom::enum_json_type child_type= child->json_type();

          add_if_missing(child, duplicates, result);

          if ((child_type == Json_dom::J_ARRAY) ||
              (child_type == Json_dom::J_OBJECT))
          {
            // now recurse and add all objects and arrays under the child
            child->find_child_doms(path_leg, true, auto_wrap,
                                   duplicates, result);
          }
        } // end of loop through children
      }

      break;
    }
  case jpl_array_cell_wildcard:
    {
      if (dom_type == Json_dom::J_ARRAY)
      {
        const Json_array * array= down_cast<const Json_array *>(this);

        for (unsigned idx= 0; idx < array->size(); idx++)
        {
          add_if_missing((*array)[idx], duplicates, result);
        }
      }

      break;
    }
  case jpl_member:
    {
      if (dom_type == Json_dom::J_OBJECT)
      {
        const Json_object * object= down_cast<const Json_object *>(this);
        std::string member_name(path_leg->get_member_name(),
                                path_leg->get_member_name_length());
        Json_dom * child= object->get(member_name);

        if (child != NULL)
          add_if_missing(child, duplicates, result);
      }

      break;
    }
  case jpl_member_wildcard:
    {
      if (dom_type == Json_dom::J_OBJECT)
      {
        const Json_object * object= down_cast<const Json_object *>(this);

        for (Json_object_iterator iter= object->iterator();
             !iter.empty();
             iter.next())
        {
          std::pair<const std::string*, Json_dom *> elt= iter.elt();
          add_if_missing(elt.second, duplicates, result);
        }
      }

      break;
    }
  default:
    DBUG_ABORT();                               /* purecov: inspected */
  }

  return;
}


Json_object::Json_object()
  : Json_dom(),
    m_map(Json_object_map::key_compare(),
          Json_object_map::allocator_type(key_memory_JSON))
{}


Json_object::~Json_object()
{
  clear();
}


/**
  This class overrides the methods on BaseReaderHandler to make
  out own handler which will construct our DOM from the parsing
  of the JSON text.
  <code>
  void Null() {   }
  void Bool(bool) {   }
  void Int(int) {   }
  void Uint(unsigned) {   }
  void Int64(int64_t) {   }
  void Uint64(uint64_t) {   }
  void Double(double) {   }
  void String(const Ch*, SizeType, bool) {   }
  void StartObject() {   }
  void EndObject(SizeType) {   }
  void StartArray() {   }
  void EndArray(SizeType) {   }
  </code>
  @see Json_dom::parse
*/
class Rapid_json_handler : public BaseReaderHandler<UTF8<> >
{
private:

// std::cerr << "callback " << name << ':' << state << '\n'; std::cerr.flush()
#define DUMP_CALLBACK(name,state)

  enum enum_state
  {
    expect_anything,
    expect_array_value,
    expect_object_key,
    expect_object_value,
    expect_eof
  };

  struct Current_element
  {
    Current_element(bool object, const char *str, uint32 length,
                    Json_dom *value)
      : m_object(object), m_key(std::string(str, length)), m_value(value)
    {}

    Current_element(Json_dom *value)
      : m_object(false), m_value(value)
    {}

    bool m_object; //!< true of object, false if array
    std::string m_key; //!< only used if object
    Json_dom *m_value; //!< deallocated by clients
  };

  struct Partial_compound
  {
    Partial_compound(bool is_object)
      : m_elements(Malloc_allocator<Current_element>(key_memory_JSON)),
        m_is_object(is_object)
    {}
    ~Partial_compound()
    {}

    std::vector<Current_element, Malloc_allocator<Current_element> > m_elements;
    bool m_is_object;
  };

  enum_state m_state;
  std::vector<Partial_compound, Malloc_allocator<Partial_compound> > m_stack;
  Json_dom *m_dom_as_built;
public:
  Rapid_json_handler()
    : m_state(expect_anything),
      m_stack(Malloc_allocator<Partial_compound>(key_memory_JSON)),
      m_dom_as_built(NULL)
  {};

  ~Rapid_json_handler()
  {
    if (m_dom_as_built)
    {
      // We managed to build something, but found garbage after it
      delete m_dom_as_built;
    }
    else
    {
      // We have something half built, empty the allocated data in it
      for (std::vector<Partial_compound, Malloc_allocator<Partial_compound> >
             ::iterator iter= m_stack.begin(); iter != m_stack.end(); ++iter)
      {
        for (std::vector<Current_element, Malloc_allocator<Current_element> >
               ::iterator i= iter->m_elements.begin();
             i != iter->m_elements.end(); ++i)
        {
          delete i->m_value;
        }
      }
    }
  }

  /**
    @returns The built JSON DOM object.
    Deallocation is the returned value responsibility of the caller.
  */
  Json_dom *get_built_doc()
  {
    Json_dom *result= m_dom_as_built;
    m_dom_as_built= NULL;
    return result;
  }

  void seeing_scalar(Json_scalar *scalar)
  {
    switch (m_state)
    {
    case expect_anything:
      m_dom_as_built= scalar;
      m_state= expect_eof;
      break;
    case expect_array_value:
      m_stack.back().m_elements.push_back(Current_element(scalar));
      break;
    case expect_object_key:
    case expect_eof:
      /* purecov: begin inspected */
      DBUG_ABORT();
      delete scalar;
      break;
      /* purecov: end inspected */
    case expect_object_value:
      DBUG_ASSERT(!m_stack.back().m_elements.empty());
      DBUG_ASSERT(m_stack.back().m_elements.back().m_value == NULL);
      m_stack.back().m_elements.back().m_value= scalar;
      m_state= expect_object_key;
      break;
    }

  }

  void Null()
  {
    DUMP_CALLBACK("null", state);
    seeing_scalar(new Json_null());
  }

  void Bool(bool b)
  {
    DUMP_CALLBACK("bool", state);
    seeing_scalar(new Json_boolean(b));
  }

  void Int(int i)
  {
    DUMP_CALLBACK("int", state);
    seeing_scalar(new Json_int(i));
  }

  void Uint(unsigned u)
  {
    DUMP_CALLBACK("unit", state);
    seeing_scalar(new Json_int(static_cast<longlong>(u)));
  }

  void Int64(int64_t i)
  {
    DUMP_CALLBACK("int64", state);
    seeing_scalar(new Json_int(i));
  }

  void Uint64(uint64_t ui64)
  {
    DUMP_CALLBACK("uint64", state);
    seeing_scalar(new Json_uint(ui64));
  }

  void Double(double d, bool is_int= false)
  {
    if (is_int)
    {
      /*
        The is_int flag is true only if -0 was seen. Handle it as an
        integer.
      */
      DBUG_ASSERT(d == 0.0);
      Int64(static_cast<int64_t>(d));
    }
    else
    {
      DUMP_CALLBACK("double", state);
      seeing_scalar(new Json_double(d));
    }
  }

  void String(const char* str, SizeType length, bool copy)
  {
    DUMP_CALLBACK("string", state);
    switch (m_state)
    {
    case expect_anything:
      m_dom_as_built= new Json_string(std::string(str, length));
      m_state= expect_eof;
      break;
    case expect_array_value:
      m_stack.back().m_elements.
        push_back(Current_element(new Json_string(std::string(str, length))));
      break;
    case expect_object_key:
      m_stack.back().m_elements.push_back(Current_element(true, str,
                                                          length, NULL));
      m_state= expect_object_value;
      break;
    case expect_eof:
      /* purecov: begin inspected */
      DBUG_ABORT();
      break;
      /* purecov: end inspected */
    case expect_object_value:
      DBUG_ASSERT(!m_stack.back().m_elements.empty());
      DBUG_ASSERT(m_stack.back().m_elements.back().m_value == NULL);
      m_stack.back().m_elements.back().m_value=
        new Json_string(std::string(str, length));
      m_state= expect_object_key;
      break;
    }
  }

  void StartObject()
  {
    DUMP_CALLBACK("start object {", state);
    switch (m_state)
    {
    case expect_anything:
    case expect_array_value:
    case expect_object_value:
      m_stack.push_back(Partial_compound(true));
      m_state= expect_object_key;
      break;
    case expect_eof:
    case expect_object_key:
      /* purecov: begin inspected */
      DBUG_ABORT();
      break;
      /* purecov: end inspected */
    }
  }

  void EndObject(SizeType)
  {
    DUMP_CALLBACK("} end object", state);
    switch (m_state)
    {
    case expect_object_key:
      {
        Json_object *o= new Json_object();
        for (std::vector<Current_element, Malloc_allocator<Current_element> >
               ::const_iterator iter= m_stack.back().m_elements.begin();
             iter != m_stack.back().m_elements.end(); ++iter)
        {
          /* _alias: save superfluous copy/delete */
          o->add_alias(iter->m_key, iter->m_value);
        }
        m_stack.pop_back();

        if (m_stack.empty())
        {
          m_dom_as_built= o;
          m_state= expect_eof;
        }
        else if (m_stack.back().m_is_object)
        {
          m_stack.back().m_elements.back().m_value= o;
          m_state= expect_object_key;
        }
        else
        {
          m_stack.back().m_elements.push_back(o);
          m_state= expect_array_value;
        }
      }
      break;
    case expect_array_value:
    case expect_eof:
    case expect_object_value:
    case expect_anything:
      /* purecov: begin inspected */
      DBUG_ABORT();
      break;
      /* purecov: end inspected */
    }
  }

  void StartArray()
  {
    DUMP_CALLBACK("start array [", state);
    switch (m_state)
    {
    case expect_anything:
    case expect_array_value:
    case expect_object_value:
      m_stack.push_back(Partial_compound(false));
      m_state= expect_array_value;
      break;
    case expect_eof:
    case expect_object_key:
      /* purecov: begin inspected */
      DBUG_ABORT();
      break;
      /* purecov: end inspected */
    }
  }

  void EndArray(SizeType)
  {
    DUMP_CALLBACK("] end array", state);
    switch (m_state)
    {
    case expect_array_value:
      {
        Json_array *o= new Json_array();
        for (std::vector<Current_element, Malloc_allocator<Current_element> >
               ::const_iterator iter= m_stack.back().m_elements.begin();
             iter != m_stack.back().m_elements.end(); ++iter)
        {
          /* _alias: save superfluous copy/delete */
          o->append_alias(iter->m_value);
        }
        m_stack.pop_back();

        if (m_stack.empty())
        {
          m_dom_as_built= o;
          m_state= expect_eof;
        }
        else
        {
          if (m_stack.back().m_is_object)
          {
            m_stack.back().m_elements.back().m_value= o;
            m_state= expect_object_key;
          }
          else
          {
            m_stack.back().m_elements.push_back(o);
            m_state= expect_array_value;
          }
        }
      }
      break;
    case expect_object_key:
    case expect_object_value:
    case expect_eof:
    case expect_anything:
      /* purecov: begin inspected */
      DBUG_ABORT();
      break;
      /* purecov: end inspected */
    }
  }

};


Json_dom *Json_dom::parse(const char *text, size_t length,
                          const char **syntaxerr, size_t *offset)
{
  Rapid_json_handler handler;
  StringStreamSafe ss(text, length);
  Reader reader;
  bool success= reader.AcceptAnyRoot().
    Parse<0, StringStreamSafe, Rapid_json_handler>(ss, handler);

  if (success)
    return handler.get_built_doc();

  // Report the error offset and the error message if requested by the caller.
  if (offset != NULL)
    *offset= reader.GetErrorOffset();
  if (syntaxerr != NULL)
    *syntaxerr= reader.GetParseError();

  return NULL;
}


/**
  Map the JSON type used by the binary representation to the type
  used by Json_dom and Json_wrapper.
  <p/>
  Note: Does not look into opaque values to determine if they
  represent decimal or date/time values. For that, look into the
  Value an retrive field_type.

  @param[in]  bintype
  @returns the JSON_dom JSON type.
*/
Json_dom::enum_json_type
Json_dom::bjson2json(const json_binary::Value::enum_type bintype)
{
  Json_dom::enum_json_type res= J_ERROR;

  switch (bintype)
  {
  case json_binary::Value::STRING:
    res= J_STRING;
    break;
  case json_binary::Value::INT:
    res= J_INT;
    break;
  case json_binary::Value::UINT:
    res= J_UINT;
    break;
  case json_binary::Value::DOUBLE:
    res= J_DOUBLE;
    break;
  case json_binary::Value::LITERAL_TRUE:
  case json_binary::Value::LITERAL_FALSE:
    res= J_BOOLEAN;
    break;
  case json_binary::Value::LITERAL_NULL:
    res= J_NULL;
    break;
  case json_binary::Value::ARRAY:
    res= J_ARRAY;
    break;
  case json_binary::Value::OBJECT:
    res= J_OBJECT;
    break;
  case json_binary::Value::ERROR:
    res= J_ERROR;
    break;
  case json_binary::Value::OPAQUE:
    res= J_OPAQUE;
    break;
  }

  return res;
}


Json_dom *Json_dom::parse(const json_binary::Value &v)
{
  Json_dom *result= NULL;

  switch (v.type())
  {
  case json_binary::Value::OBJECT:
    {
      Json_object *jo= new Json_object();
      for (uint32 i= 0; i < v.element_count(); ++i)
      {
        jo->add_alias(std::string(v.key(i).get_data(),
                                  v.key(i).get_data_length()),
                      parse(v.element(i)));
      }
      result= jo;
      break;
    }
  case json_binary::Value::ARRAY:
    {
      Json_array *jarr= new Json_array();
      for (uint32 i= 0; i < v.element_count(); ++i)
      {
        jarr->append_alias(parse(v.element(i)));
      }

      result= jarr;
      break;
    }
  case json_binary::Value::DOUBLE:
    result= new Json_double(v.get_double());
    break;
  case json_binary::Value::INT:
    result= new Json_int(v.get_int64());
    break;
  case json_binary::Value::UINT:
    result= new Json_uint(v.get_uint64());
    break;
  case json_binary::Value::LITERAL_FALSE:
    result= new Json_boolean(false);
    break;
  case json_binary::Value::LITERAL_TRUE:
    result= new Json_boolean(true);
    break;
  case json_binary::Value::LITERAL_NULL:
    result= new Json_null();
    break;
  case json_binary::Value::OPAQUE:
    {
      const enum_field_types ftyp= v.field_type();

      if (ftyp == MYSQL_TYPE_NEWDECIMAL)
      {
        my_decimal m;
        if (Json_decimal::convert_from_binary(v.get_data(),
                                              v.get_data_length(),
                                              &m))
          my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
        result= new Json_decimal(m);
      }
      else if (ftyp == MYSQL_TYPE_DATE ||
               ftyp == MYSQL_TYPE_TIME ||
               ftyp == MYSQL_TYPE_DATETIME)
      {
        MYSQL_TIME t;
        Json_datetime::from_packed(v.get_data(), ftyp, &t);
        result= new Json_datetime(t, ftyp);
      }
      else
      {
        result= new Json_opaque(v.field_type(),
                                v.get_data(),
                                v.get_data_length());
      }
      break;
    }
  case json_binary::Value::STRING:
    result= new Json_string(std::string(v.get_data(), v.get_data_length()));
    break;

  case json_binary::Value::ERROR:
    {
      /* purecov: begin inspected */
      DBUG_ABORT();
      my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
      break;
      /* purecov: end inspected */
    }
  }

  return result;
}

void Json_dom::replace_dom_in_container(Json_dom *oldv, Json_dom *newv)
{
  switch (json_type())
  {
  case J_ARRAY:
    {
      Json_array * const thisa= down_cast<Json_array *>(this);

      for (size_t i=0; i <= thisa->size() - 1; ++i)
      {
        if ((*thisa)[i] == oldv)
        {
          thisa->replace_alias(i, newv);
          break;
        }
      }
      break;
    }
  case J_OBJECT:
    {
      Json_object * const thiso= down_cast<Json_object *>(this);

      for (Json_object_iterator it= thiso->iterator(); !it.empty(); it.next())
      {
        std::pair<const std::string*, Json_dom *> e= it.elt();

        if (e.second == oldv)
        {
          std::string eltnam= *e.first;
          thiso->remove(*e.first);
          thiso->add_alias(eltnam, newv);
          break;
        }
      }
      break;
    }
  default:
    DBUG_ABORT();                               /* purecov: inspected */
  }
}

bool Json_object::add_clone(const std::string &key, const Json_dom *value)
{
  if (m_map.find(key) != m_map.end())
  {
    /*
      We have already an element with this key.  Note we compare utf-8 bytes
      directly here. It's complicated when when you take into account composed
      and decomposed forms of accented characters and ligatures: different
      sequences might encode the same glyphs but we ignore that for now.  For
      example, the code point U+006E (the Latin lowercase "n") followed by
      U+0303 (the combining tilde) is defined by Unicode to be canonically
      equivalent to the single code point U+00F1 (the lowercase letter of the
      Spanish alphabet).  For now, users must normalize themselves to avoid
      element dups.

      This is what ECMAscript does also: "Two IdentifierName that are
      canonically equivalent according to the Unicode standard are not equal
      unless they are represented by the exact same sequence of code units (in
      other words, conforming ECMAScript implementations are only required to
      do bitwise comparison on IdentifierName values). The intent is that the
      incoming source text has been converted to normalised form C before it
      reaches the compiler." (ECMA-262 5.1 edition June 2011)

      See WL-2048 Add function for Unicode normalization
    */
    return false;
  }
  Json_dom *v= value->clone();

#ifndef DBUG_OFF
  std::pair<Json_object_map::const_iterator, bool> ret=
#endif
    m_map.insert(std::make_pair(key, v));
  DBUG_ASSERT(ret.second);

  v->set_parent(this);

  return true;
}


bool Json_object::add_alias(const std::string &key, Json_dom *value)
{
  if (m_map.find(key) != m_map.end())
  {
    // we have already an element with this key
    delete value; // our responsibility now
    return false;
  }
#ifndef DBUG_OFF
  std::pair<Json_object_map::const_iterator, bool> ret=
#endif
    m_map.insert(std::make_pair(key, value));
  DBUG_ASSERT(ret.second);

  value->set_parent(this);

  return true;
}

void Json_object::consume(Json_object *other)
{
  Json_object_map this_map= m_map;
  Json_object_map other_map= other->m_map;

  for (Json_object_map::iterator other_iter= other_map.begin();
       other_iter != other_map.end(); ++other_iter)
  {
    std::string key= other_iter->first;
    Json_dom *value= other_iter->second->clone();

    Json_object_map::iterator this_iter= this_map.find(key);

    if (this_iter != this_map.end())
    {
      /*
        Oops. Duplicate key. Merge the values.
        This is where the recursion in JSN_MERGE() occurs.
      */
      Json_dom *original_value= this_iter->second;
      value= merge_doms(original_value->clone(), value);

      remove(key);
    }

    add_alias(key, value);
  }

  delete other;
}

Json_dom *Json_object::get(const std::string key) const
{
  const Json_object_map::const_iterator iter= m_map.find(key);

  if (iter != m_map.end())
    return iter->second;

  return NULL;
}


bool Json_object::remove(const std::string key)
{
  const Json_object_map::iterator iter= m_map.find(key);

  if (iter != m_map.end())
  {
    delete iter->second;
    m_map.erase(iter);
    return true;
  }

  return false;
}


bool Json_object::remove(const Json_dom *child)
{
  for (Json_object_map::iterator iter= m_map.begin();
       iter != m_map.end(); ++iter)
  {
    Json_dom *candidate= iter->second;

    if (child == candidate)
    {
      delete candidate;
      m_map.erase(iter);
      return true;
    }
  } // end of loop through children

  return false;
}


size_t Json_object::cardinality() const
{
  return m_map.size();
}


uint32 Json_object::depth() const
{
  uint deepest_child= 0;

  for (Json_object_map::const_iterator iter= m_map.begin();
       iter != m_map.end(); ++iter)
  {
    deepest_child= std::max(deepest_child, iter->second->depth());
  }
  return 1 + deepest_child;
}


Json_dom *Json_object::clone() const
{
  Json_object * const o= new Json_object();

  for (Json_object_map::const_iterator iter= m_map.begin();
       iter != m_map.end(); ++iter)
  {
    o->add_clone(iter->first, iter->second);
  }
  return o;
}


void Json_object::clear()
{
  for (Json_object_map::const_iterator iter= m_map.begin();
       iter != m_map.end(); ++iter)
  {
    delete iter->second;
  }
  m_map.clear();
}


Json_object_iterator Json_object::iterator() const
{
  return Json_object_iterator(m_map);
}


/**
  Compare two keys from a JSON object and determine whether or not the
  first key is less than the second key. key1 is considered less than
  key2 if

  a) key1 is shorter than key2, or if

  b) key1 and key2 have the same length, but different contents, and
  the first byte that differs has a smaller value in key1 than in key2

  Otherwise, key1 is not less than key2.

  @param key1 the first key to compare
  @param key2 the second key to compare
  @return true if key1 is considered less than key2, false otherwise
*/
bool Json_key_comparator::operator() (const std::string &key1,
                                      const std::string &key2) const
{
  if (key1.length() != key2.length())
    return key1.length() < key2.length();

  return memcmp(key1.data(), key2.data(), key1.length()) < 0;
}


Json_array::Json_array()
  : Json_dom(), m_v(Json_dom_vector::allocator_type(key_memory_JSON))
{}


Json_array::Json_array(Json_dom *innards)
  : Json_dom(), m_v(Json_dom_vector::allocator_type(key_memory_JSON))
{
  append_alias(innards);
}


Json_array::~Json_array()
{
  delete_container_pointers(m_v);
}


void Json_array::append_clone(const Json_dom *value)
{
  Json_dom *clone= value->clone();
  m_v.push_back(clone);
  clone->set_parent(this);
}


void Json_array::append_alias(Json_dom *value)
{
  m_v.push_back(value);
  value->set_parent(this);
}


void Json_array::consume(Json_array *other)
{
  Json_dom_vector other_vector= other->m_v;

  for (Json_dom_vector::iterator iter= other_vector.begin();
       iter != other_vector.end(); ++iter)
  {
    append_clone(*iter);
  }

  delete other;
}


void Json_array::insert_clone(size_t index, const Json_dom *value)
{
  Json_dom_vector::iterator iter= m_v.begin();

  Json_dom *insertable;

  if (index < m_v.size())
  {
    insertable=  value->clone();
    m_v.insert(iter + index, insertable);
  }
  else
  {
    //append needed
    insertable=  value->clone();
    m_v.push_back(insertable);
  }

  insertable->set_parent(this);
}


void Json_array::insert_alias(size_t index, Json_dom *value)
{
  Json_dom_vector::iterator iter= m_v.begin();

  if (index < m_v.size())
  {
    m_v.insert(iter + index, value);
  }
  else
  {
    //append needed
    m_v.push_back(value);
  }

  value->set_parent(this);
}


bool Json_array::remove(size_t index)
{
  if (index >= m_v.size())
  {
    return false;
  }

  const Json_dom_vector::iterator iter= m_v.begin() + index;

  if (iter != m_v.end())
  {
    delete *iter;
    m_v.erase(iter);
    return true;
  }

  return false;
}


bool Json_array::remove(const Json_dom *child)
{
  for (Json_dom_vector::iterator iter= m_v.begin(); iter != m_v.end(); ++iter)
  {
    Json_dom *candidate= *iter;

    if (child == candidate)
    {
      delete candidate;
      m_v.erase(iter);
      return true;
    }
  } // end of loop through children

  return false;
}


uint32 Json_array::depth() const
{
  uint deepest_child= 0;

  for (Json_dom_vector::const_iterator it= m_v.begin(); it != m_v.end(); ++it)
  {
    deepest_child= std::max(deepest_child, (*it)->depth());
  }
  return 1 + deepest_child;
}


Json_dom *Json_array::clone() const
{
  Json_array * const vv= new Json_array();

  for (Json_dom_vector::const_iterator it= m_v.begin(); it != m_v.end(); ++it)
  {
    vv->append_clone(*it);
  }
  return vv;
}


void Json_array::replace_clone(size_t index, const Json_dom *new_value)
{
  delete m_v[index];
  Json_dom *clone= new_value->clone();
  m_v[index]= clone;
  clone->set_parent(this);
}


void Json_array::replace_alias(size_t index, Json_dom *new_value)
{
  delete m_v[index];
  m_v[index]= new_value;
  new_value->set_parent(this);
}


void Json_array::clear()
{
  delete_container_pointers(m_v);
}

/**
  Perform quoting on a JSON string to make an external representation
  of it. it wraps double quotes (text quotes) around the string (cptr)
  an also performs escaping according to the following table:
  <pre>
  Common name     C-style  Original unescaped     Transformed to
                  escape   UTF-8 bytes            escape sequence
                  notation                        in UTF-8 bytes
  ---------------------------------------------------------------
  quote           \"       %x22                    %x5C %x22
  backslash       \\       %x5C                    %x5C %x5C
  backspace       \b       %x08                    %x5C %x62
  formfeed        \f       %x0C                    %x5C %x66
  linefeed        \n       %x0A                    %x5C %x6E
  carriage-return \r       %x0D                    %x5C %x72
  tab             \t       %x09                    %x5C %x74
  unicode         \uXXXX  A hex number in the      %x5C %x75
                          range of 00-1F,          followed by
                          except for the ones      4 hex digits
                          handled above (backspace,
                          formfeed, linefeed,
                          carriage-return,
                          and tab).
  ---------------------------------------------------------------
  </pre>
*/
std::string double_quote(const char *cptr, size_t length)
{
  std::stringbuf buf;
  buf.sputc('"');
  for (size_t i= 0; i < length; i++)
  {
    char esc[2]= {'\\', cptr[i]};
    bool done= true;
    switch (cptr[i])
    {
    case '"' :
    case '\\' :
      break;
    case '\b':
      esc[1]= 'b';
      break;
    case '\f':
      esc[1]= 'f';
      break;
    case '\n':
      esc[1]= 'n';
      break;
    case '\r':
      esc[1]= 'r';
      break;
    case '\t':
      esc[1]= 't';
      break;
    default:
      done= false;
    }

    if (done)
    {
      buf.sputc(esc[0]);
      buf.sputc(esc[1]);
    }
    else if (((cptr[i] & ~0x7f) == 0) && // bit 8 not set
             (cptr[i] < 0x1f))
    {
      /* 
        Unprintable control character, use hex a hexadecimal number.
        The meaning of such a number determined by ISO/IEC 10646.
      */
      buf.sputn("\\u00", 4);
      std::ostringstream s;
      s << std::hex << (256 + cptr[i]);  // add 256 for leading zero
      buf.sputn(s.str().c_str() + 1, 2);
    }
    else
    {
      buf.sputc(cptr[i]);
    }
  }
  buf.sputc('"');
  return buf.str();
}


Json_decimal::Json_decimal(const my_decimal &value)
  : Json_number(), m_dec(value)
{}


int Json_decimal::binary_size() const
{
  /*
    We need two bytes for the precision and the scale, plus whatever
    my_decimal2binary() needs.
  */
  return 2 + my_decimal_get_binary_size(m_dec.precision(), m_dec.frac);
}


bool Json_decimal::get_binary(char* dest) const
{
  DBUG_ASSERT(binary_size() <= MAX_BINARY_SIZE);
  /*
    my_decimal2binary() loses the precision and the scale, so store them
    in the first two bytes.
  */
  dest[0]= static_cast<char>(m_dec.precision());
  dest[1]= static_cast<char>(m_dec.frac);
  // Then store the decimal value.
  return my_decimal2binary(E_DEC_ERROR, &m_dec,
                           pointer_cast<uchar*>(dest) + 2,
                           m_dec.precision(), m_dec.frac) != E_DEC_OK;
}


bool Json_decimal::convert_from_binary(const char *bin, size_t len,
                                       my_decimal *dec)
{
  if (len < 2)
    return true;

  int precision= bin[0];
  int scale= bin[1];

  size_t bin_size= my_decimal_get_binary_size(precision, scale);
  if (bin_size != len - 2)
    return true;

  return binary2my_decimal(E_DEC_ERROR, pointer_cast<const uchar*>(bin) + 2,
                           dec, precision, scale) != E_DEC_OK;
}


Json_dom *Json_double::clone() const
{
  return new Json_double(m_f);
}


Json_dom::enum_json_type Json_datetime::json_type () const
{
  switch (m_field_type)
  {
  case MYSQL_TYPE_TIME : return J_TIME;
  case MYSQL_TYPE_DATETIME: return J_DATETIME;
  case MYSQL_TYPE_DATE: return J_DATE;
  case MYSQL_TYPE_TIMESTAMP: return J_TIMESTAMP;
  default: ;
  }
  /* purecov: begin inspected */
  DBUG_ABORT();
  return J_NULL;
  /* purecov: end inspected */
}


Json_dom *Json_datetime::clone() const
{
  return new Json_datetime(m_t, m_field_type);
}


void Json_datetime::to_packed(char *dest) const
{
  longlong packed= TIME_to_longlong_packed(&m_t);
  int8store(dest, packed);
}


void Json_datetime::from_packed(const char *from, enum_field_types ft,
                                MYSQL_TIME *to)
{
  TIME_from_longlong_packed(to, ft, sint8korr(from));
}


Json_opaque::Json_opaque(enum_field_types mytype, const char *v, size_t size)
  : Json_scalar(), m_mytype(mytype), m_val(v, size)
{}


Json_dom *Json_opaque::clone() const
{
  return new Json_opaque(m_mytype, value(), size());
}


Json_object_iterator::Json_object_iterator(const Json_object_map &map)
  : m_end(map.end())
{
  m_iter= map.begin();
}


bool Json_object_iterator::empty() const
{
  return m_iter == m_end;
}


void Json_object_iterator::next()
{
  m_iter++;
}


std::pair<const std::string*, Json_dom *> Json_object_iterator::elt() const
{
  return std::make_pair(&m_iter->first, m_iter->second);
}


Json_wrapper_object_iterator::
Json_wrapper_object_iterator(const Json_object_map &map)
  : m_is_dom(true), m_iter(map.begin()), m_end(map.end()), m_element_count(-1)
{
  m_iter= map.begin();
}


Json_wrapper_object_iterator::
Json_wrapper_object_iterator(const json_binary::Value *value)
  : m_is_dom(false), m_element_count(value->element_count()), m_value(value)
{
  m_curr_element= 0;
}


bool Json_wrapper_object_iterator::empty() const
{
  return m_is_dom ? (m_iter == m_end) : (m_curr_element >= m_element_count);
}


void Json_wrapper_object_iterator::next()
{
  if (m_is_dom)
  {
    m_iter++;
  }
  else
  {
    ++m_curr_element;
  }
}


std::pair<const std::string, Json_wrapper>
Json_wrapper_object_iterator::elt() const
{
  if (m_is_dom)
  {
    Json_wrapper wr(m_iter->second);
    // DOM possibly owned by object and we don't want to make a clone
    wr.set_alias();
    return std::make_pair(m_iter->first, wr);
  }

  std::string key(m_value->key(m_curr_element).get_data(),
                  m_value->key(m_curr_element).get_data_length());
  Json_wrapper wr(m_value->element(m_curr_element));
  return std::make_pair(key, wr);
}


Json_wrapper::Json_wrapper(Json_dom *dom_value) :
  m_is_dom(true), m_dom_alias(false), m_value(),
  m_id(NULL), m_dom_value(dom_value)
{
  if (!dom_value)
  {
    m_dom_alias= true; //!< no deallocation, make us empty
  }
}


void Json_wrapper::steal(Json_wrapper *old)
{
  if (old->m_is_dom)
  {
    bool old_is_aliased= old->m_dom_alias;
    old->m_dom_alias= true; // we want no deep copy now, or later
    *this= *old;
    this->m_dom_alias= old_is_aliased; // set it back
    // old is now marked as aliased, so any ownership is effectively
    // transferred to this.
  }
  else
  {
    *this= *old;
  }
}

Json_wrapper::Json_wrapper(const json_binary::Value &value) :
  m_is_dom(false), m_dom_alias(false), m_value(value),
  m_id(NULL), m_dom_value(NULL)
{}


Json_wrapper::Json_wrapper(const json_binary::Value &value, const char *id) :
  m_is_dom(false), m_dom_alias(false), m_value(value),
  m_id(id), m_dom_value(NULL)
{}


Json_wrapper::Json_wrapper(const Json_wrapper &old) :
  m_is_dom(old.m_is_dom),
  m_dom_alias(old.m_dom_alias),
  m_value(old.m_value),
  m_id(old.m_id),
  m_dom_value(old.m_is_dom ?
              (m_dom_alias? old.m_dom_value : old.m_dom_value->clone()) :
              NULL),
  m_tmp(old.m_tmp)
{}


Json_wrapper::~Json_wrapper()
{
  if (m_is_dom && !m_dom_alias)
  {
    // we own our own copy, so we are responsible for deallocation
    delete m_dom_value;
  }
}


Json_wrapper &Json_wrapper::operator=(const Json_wrapper& from)
{
  if (this == &from)
  {
    return *this;   // self assignment: no-op
  }

  if (m_is_dom && !m_dom_alias &&!empty())
  {
    // we own our own copy, so we are responsible for deallocation
    delete m_dom_value;
  }

  m_is_dom= from.m_is_dom;

  if (from.m_is_dom)
  {
    if (from.m_dom_alias)
    {
      m_dom_value= from.m_dom_value;
    }
    else
    {
      m_dom_value= from.m_dom_value->clone();
    }

    m_dom_alias= from.m_dom_alias;
  }
  else
  {
    m_dom_value= NULL;
    m_value= from.m_value;
    m_id= from.m_id;
  }

  m_tmp= from.m_tmp;
  return *this;
}


Json_dom *Json_wrapper::to_dom()
{
  if (!m_is_dom)
  {
    // Build a DOM from the binary JSON value and
    // convert this wrapper to hold the DOM instead
    m_dom_value= Json_dom::parse(m_value);
    m_is_dom= true;
    m_dom_alias= false;
  }

  return m_dom_value;
}


json_binary::Value Json_wrapper::to_value()
{
  if (empty())
  {
    return json_binary::Value();
  }

  if (m_is_dom)
  {
    switch (json_binary::serialize(m_dom_value, &m_tmp))
    {
    case json_binary::OK:
      return json_binary::parse_binary(m_tmp.ptr(), m_tmp.length());
    case json_binary::VALUE_TOO_BIG:
      my_error(ER_JSON_VALUE_TOO_BIG, MYF(0));
      break;
    case json_binary::KEY_TOO_BIG:
      my_error(ER_JSON_KEY_TOO_BIG, MYF(0));
      break;
    case json_binary::FAILED:
      my_error(ER_INVALID_JSON_TEXT, MYF(0));
      break;
    case json_binary::OOM:
      my_error(ER_OUT_OF_RESOURCES, MYF(0));
      break;
    }

    // An error occurred, return an empty value.
    return json_binary::Value();
  }

  return m_value;
}


inline void single_quote(String *buffer, bool json_quoted)
{
  if (json_quoted)
    buffer->append('"');
}

bool Json_wrapper::to_string(String *buffer, bool json_quoted)
{
  buffer->set_charset(&my_charset_utf8mb4_bin);

  if (empty())
  {
    buffer->append("NULL");
    return true;
  }

  switch (type())
  {
  case Json_dom::J_TIME:
  case Json_dom::J_DATE:
  case Json_dom::J_DATETIME:
  case Json_dom::J_TIMESTAMP:
    {
      MYSQL_TIME t;
      get_datetime(&t);
      single_quote(buffer, json_quoted);
      char dtbuff[MAX_DATE_STRING_REP_LENGTH];
      const int size= my_TIME_to_str(&t, dtbuff, 6);
      buffer->append(dtbuff, size);
      single_quote(buffer, json_quoted);
      break;
    }
  case Json_dom::J_ARRAY:
    {
      buffer->append('[');

      for (uint32 i= 0; i < length(); ++i)
      {
        if (i > 0)
        {
          buffer->append(", ");
        }

        (*this)[i].to_string(buffer, json_quoted);
      }
      buffer->append(']');
      break;
    }
  case Json_dom::J_BOOLEAN:
    buffer->append(get_boolean() ? "true" : "false");
    break;
  case Json_dom::J_DECIMAL:
    {
      char buff[DECIMAL_MAX_STR_LENGTH+1];
      int length= DECIMAL_MAX_STR_LENGTH + 1;
      my_decimal m;
      if (get_decimal_data(&m) ||
          decimal2string(&m, buff, &length, 0, 0, 0))
        my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
      else
        buffer->append(buff);
      break;
    }
  case Json_dom::J_DOUBLE:
    {
      double d= get_double();
      char to[FLOATING_POINT_BUFFER];
      size_t len= my_gcvt(d, MY_GCVT_ARG_DOUBLE,
                          FLOATING_POINT_BUFFER - 1, to, NULL);
      buffer->append(to, len);
      break;
    }
  case Json_dom::J_INT:
    {
      buffer->append_longlong(get_int());
      break;
    }
  case Json_dom::J_NULL:
    buffer->append("null");
    break;
  case Json_dom::J_OBJECT:
    {
      buffer->append('{');
      uint32 i= 0;
      for (Json_wrapper_object_iterator iter= object_iterator();
           !iter.empty(); iter.next())
      {
        if (i++ > 0)
        {
          buffer->append(", ");
        }
        buffer->append('"');
        buffer->append(iter.elt().first.c_str());
        buffer->append("\": ");
        iter.elt().second.to_string(buffer, true);
      }
      buffer->append('}');

      break;
    }
  case Json_dom::J_OPAQUE:
    {
      single_quote(buffer, json_quoted);
      buffer->append("base64:type", 11);

      buffer->append_ulonglong(field_type());
      buffer->append(':');

      if (get_data_length() > (uint32)base64_encode_max_arg_length())
      {
        buffer->append("<data too long to decode - unexpected error>\"");
        my_error(ER_INTERNAL_ERROR, MYF(0),
                 "JSON: could not decode opaque data");
        return false;
      }

      size_t pos= buffer->length();
      const int needed=
        base64_needed_encoded_length(static_cast<int>(get_data_length()));
      buffer->mem_realloc(pos + needed);
      base64_encode(get_data(), get_data_length(),
                    const_cast<char*>(buffer->ptr() + pos));
      // "base64:typeXX:<binary data>"
      buffer->length(pos + needed - 1); // drop zero terminator space
      single_quote(buffer, json_quoted);
      break;
    }
  case Json_dom::J_STRING:
    {
      const char *data= get_data();
      std::string q= (json_quoted ? double_quote(data, get_data_length()) :
                      std::string(data, get_data_length()));
      buffer->append(q.c_str(), q.length());
      break;
    }
  case Json_dom::J_UINT:
    {
      buffer->append_ulonglong(get_uint());
      break;
    }
  default:
    /* purecov: begin inspected */
    DBUG_ABORT();
    my_error(ER_INTERNAL_ERROR, MYF(0), "JSON wrapper: unexpected type");
    return false;
    /* purecov: end inspected */
  }
  return true;
}


Json_dom::enum_json_type Json_wrapper::type() const
{
  if (empty())
  {
    return Json_dom::J_ERROR;
  }

  if (m_is_dom)
  {
    return m_dom_value->json_type();
  }

  json_binary::Value::enum_type typ= m_value.type();

  if (typ == json_binary::Value::OPAQUE)
  {
    const enum_field_types ftyp= m_value.field_type();

    switch (ftyp)
    {
    case MYSQL_TYPE_NEWDECIMAL:
      return Json_dom::J_DECIMAL;
    case MYSQL_TYPE_DATETIME:
      return Json_dom::J_DATETIME;
    case MYSQL_TYPE_DATE:
      return Json_dom::J_DATE;
    case MYSQL_TYPE_TIME:
      return Json_dom::J_TIME;
    case MYSQL_TYPE_TIMESTAMP:
      return Json_dom::J_TIMESTAMP;
    default: ;
      // ok, fall through
    }
  }

  return Json_dom::bjson2json(typ);
}


enum_field_types Json_wrapper::field_type() const
{
  if (m_is_dom)
  {
    return static_cast<Json_opaque *>(m_dom_value)->type();
  }

  return m_value.field_type();
}


Json_wrapper_object_iterator Json_wrapper::object_iterator() const
{
  DBUG_ASSERT(type() == Json_dom::J_OBJECT);

  if (m_is_dom)
  {
    const Json_object *o= down_cast<const Json_object *>(m_dom_value);
    return Json_wrapper_object_iterator(o->m_map);
  }

  return Json_wrapper_object_iterator(&m_value);
}


Json_wrapper Json_wrapper::operator[](size_t index) const
{
  DBUG_ASSERT(type() == Json_dom::J_ARRAY);
  if (m_is_dom)
  {
    const Json_array *o= down_cast<const Json_array *>(m_dom_value);
    Json_wrapper wr((*o)[index]);
    wr.set_alias(); // wr doesn't own the supplied DOM: part of array DOM
    return wr;
  }

  return Json_wrapper(m_value.element(index));
}


const char *Json_wrapper::get_data() const
{
  if (m_is_dom)
  {
    return type() == Json_dom::J_STRING ?
      down_cast<Json_string *>(m_dom_value)->value().c_str() :
      down_cast<Json_opaque *>(m_dom_value)->value();
  }

  return m_value.get_data();
}


size_t Json_wrapper::get_data_length() const
{
  if (m_is_dom)
  {
    return type() == Json_dom::J_STRING ?
      down_cast<Json_string *>(m_dom_value)->size() :
      down_cast<Json_opaque *>(m_dom_value)->size();
  }

  return m_value.get_data_length();
}


bool Json_wrapper::get_decimal_data(my_decimal *d) const
{
  if (m_is_dom)
  {
    *d= *static_cast<Json_decimal *>(m_dom_value)->value();
    return false;
  }

  return Json_decimal::convert_from_binary(m_value.get_data(),
                                           m_value.get_data_length(),
                                           d);
}


double Json_wrapper::get_double() const
{
  if (m_is_dom)
  {
    return static_cast<Json_double *>(m_dom_value)->value();
  }

  return m_value.get_double();
}


longlong Json_wrapper::get_int() const
{
  if (m_is_dom)
  {
    return static_cast<Json_int *>(m_dom_value)->value();
  }

  return m_value.get_int64();
}


ulonglong Json_wrapper::get_uint() const
{
  if (m_is_dom)
  {
    return static_cast<Json_uint *>(m_dom_value)->value();
  }

  return m_value.get_uint64();
}


void Json_wrapper::get_datetime(MYSQL_TIME *t) const
{
  enum_field_types ftyp= MYSQL_TYPE_NULL;

  switch(type())
  {
  case Json_dom::J_DATE:
    ftyp= MYSQL_TYPE_DATE;
    break;
  case Json_dom::J_DATETIME:
  case Json_dom::J_TIMESTAMP:
    ftyp= MYSQL_TYPE_DATETIME;
    break;
  case Json_dom::J_TIME:
    ftyp= MYSQL_TYPE_TIME;
    break;
  default:
    DBUG_ABORT();                               /* purecov: inspected */
  }

  if (m_is_dom)
  {
    *t= *static_cast<Json_datetime *>(m_dom_value)->value();
  }
  else
  {
    Json_datetime::from_packed(m_value.get_data(), ftyp, t);
  }
}


bool Json_wrapper::get_boolean() const
{
  if (m_is_dom)
  {
    return static_cast<Json_boolean *>(m_dom_value)->value();
  }

  return m_value.type() == json_binary::Value::LITERAL_TRUE;
}


Json_path Json_dom::get_location()
{
  if (m_parent == NULL)
  {
    Json_path result;
    return result;
  }

  Json_path result= m_parent->get_location();

  if (m_parent->json_type() == Json_dom::J_OBJECT)
  {
    Json_object *object= down_cast<Json_object *>(m_parent);
    for (Json_object_iterator iter= object->iterator();
         !iter.empty();
         iter.next())
    {
      std::pair<const std::string*, Json_dom *> elt= iter.elt();
      if (elt.second == this)
      {
        const std::string *key= elt.first;
        Json_path_leg child_leg(*key);
        result.append(child_leg);
        break;
      }
    }
  }
  else
  {
    DBUG_ASSERT(m_parent->json_type() == Json_dom::J_ARRAY);
    Json_array *array= down_cast<Json_array *>(m_parent);

    for (size_t idx= 0; idx < array->size(); idx++)
    {
      if ((*array)[idx] == this)
      {
        Json_path_leg child_leg(idx);
        result.append(child_leg);
        break;
      }
    }
  }

  return result;
}


void Json_dom::seek(const Json_path &path, Json_dom_vector *hits,
                    bool auto_wrap)
{
  Json_dom_vector doms((Json_dom_vector::allocator_type(key_memory_JSON)));
  Json_dom_vector
    candidates((Json_dom_vector::allocator_type(key_memory_JSON)));
  doms.push_back(this);
  Json_dom_set duplicates((Json_dom_set::key_compare()),
                          Json_dom_set::allocator_type(key_memory_JSON));

  for (size_t path_idx= 0; path_idx < path.leg_count(); path_idx++)
  {
    const Json_path_leg *path_leg= path.get_leg_at(path_idx);
    duplicates.clear();
    candidates.clear();

    for (Json_dom_vector::iterator it= doms.begin(); it != doms.end(); ++it)
    {
      (*it)->find_child_doms(path_leg, false, auto_wrap,
                             &duplicates, &candidates);
    }

    // swap the two lists so that they can be re-used
    std::swap(doms, candidates);
  }

  for (Json_dom_vector::iterator it= doms.begin(); it != doms.end(); ++it)
  {
    hits->push_back(*it);
  }
}


void Json_wrapper::seek(const Json_path &path,
                        Json_wrapper_vector *hits,
                        bool auto_wrap) const
{
  if (empty())
  {
    return;
  }

  /*
    FIXME.

    Materialize the dom if the path contains ellipses. This can
    be removed when we fix duplicate detection in
    json_binary::Value.seek().
   */
  if (((Json_path)path).contains_ellipsis())
  {
    ((Json_wrapper *) this)->to_dom();
  }

  if (m_is_dom)
  {
    Json_dom_vector dhits((Json_dom_vector::allocator_type(key_memory_JSON)));
    m_dom_value->seek(path, &dhits, auto_wrap);
    for (Json_dom_vector::iterator it= dhits.begin(); it != dhits.end(); ++it)
    {
      hits->push_back(Json_wrapper((*it)->clone()));
    }
  }
  else
  {
    json_binary::Value_vector
      values((json_binary::Value_vector::allocator_type(key_memory_JSON)));
    m_value.seek(path, auto_wrap, &values);

    for (json_binary::Value_vector::iterator it= values.begin();
         it != values.end(); ++it)
    {
      hits->push_back(Json_wrapper(*it));
    }
  }
}


size_t Json_wrapper::length() const
{
  if (empty())
  {
    return 0;
  }

  if (m_is_dom)
  {
    switch(m_dom_value->json_type())
    {
    case Json_dom::J_ARRAY:
      return down_cast<Json_array *>(m_dom_value)->size();
    case Json_dom::J_OBJECT:
      return down_cast<Json_object *>(m_dom_value)->cardinality();
    default:
      return 1;
    }
  }

  switch(m_value.type())
  {
  case json_binary::Value::ARRAY:
  case json_binary::Value::OBJECT:
    return m_value.element_count();
  default:
    return 1;
  }
}


size_t Json_wrapper::depth() const
{
  if (empty())
  {
    return 0;
  }

  if (m_is_dom)
  {
    return m_dom_value->depth();
  }

  Json_dom *d= Json_dom::parse(m_value);
  size_t result= d->depth();
  delete d;
  return result;
}


/**
  Compare two numbers of the same type.
  @param val1 the first number
  @param val2 the second number
  @retval -1 if val1 is less than val2,
  @retval 0 if val1 is equal to val2,
  @retval 1 if val1 is greater than val2
*/
template <class T> static int compare_numbers(T val1, T val2)
{
  return (val1 < val2) ? -1 : ((val1 == val2) ? 0 : 1);
}


/**
  Compare a decimal value to a double by converting the double to a
  decimal.
  @param a the decimal value
  @param b the double value
  @return -1 if a is less than b,
          0 if a is equal to b,
          1 if a is greater than b
*/
static int compare_json_decimal_double(const my_decimal &a, double b)
{
  /*
    First check the sign of the two values. If they differ, the
    negative value is the smaller one.
  */
  const bool a_is_negative= a.sign();
  const bool b_is_negative= (b < 0);
  if (a_is_negative != b_is_negative)
    return a_is_negative ? -1 : 1;

  my_decimal b_dec;
  switch (double2decimal(b, &b_dec))
  {
  case E_DEC_OK:
    return my_decimal_cmp(&a, &b_dec);
  case E_DEC_OVERFLOW:
    /*
      b is too big to fit in a DECIMAL, so it must have a
      larger absolute value than a, which is a DECIMAL.
    */
    return a_is_negative ? 1 : -1;
  case E_DEC_TRUNCATED:
    /*
      b was truncated to fit in a DECIMAL, which means that b_dec is
      closer to zero than b.
    */
    {
      int cmp= my_decimal_cmp(&a, &b_dec);

      /*
        If the truncated b_dec is equal to a, a must be closer to zero
        than b.
      */
      if (cmp == 0)
        return a_is_negative ? 1 : -1;

      return cmp;
    }
  default:
    /*
      double2decimal() is not supposed to return anything other than
      E_DEC_OK, E_DEC_OVERFLOW or E_DEC_TRUNCATED, so this should
      never happen.
    */
    DBUG_ABORT();                             /* purecov: inspected */
    return 1;                                 /* purecov: inspected */
  }
}


/**
  Compare a JSON double to a JSON signed integer.
  @param a the double value
  @param b the integer value
  @return -1 if a is less than b,
          0 if a is equal to b,
          1 if a is greater than b
*/
static int compare_json_double_int(double a, longlong b)
{
  double b_double= static_cast<double>(b);
  if (a < b_double)
    return -1;
  if (a > b_double)
    return 1;

  /*
    The two numbers were equal when compared as double. Since
    conversion from longlong to double isn't lossless, they could
    still be different. Convert to decimal to compare their exact
    values.
  */
  my_decimal b_dec;
  longlong2decimal(b, &b_dec);
  return -compare_json_decimal_double(b_dec, a);
}


/**
  Compare a JSON double to a JSON unsigned integer.
  @param a the double value
  @param b the unsigned integer value
  @return -1 if a is less than b,
          0 if a is equal to b,
          1 if a is greater than b
*/
static int compare_json_double_uint(double a, ulonglong b)
{
  double b_double= ulonglong2double(b);
  if (a < b_double)
    return -1;
  if (a > b_double)
    return 1;

  /*
    The two numbers were equal when compared as double. Since
    conversion from longlong to double isn't lossless, they could
    still be different. Convert to decimal to compare their exact
    values.
  */
  my_decimal b_dec;
  ulonglong2decimal(b, &b_dec);
  return -compare_json_decimal_double(b_dec, a);
}


/**
  Compare a JSON signed integer to a JSON unsigned integer.
  @param a the signed integer
  @param b the unsigned integer
  @return -1 if a is less than b,
          0 if a is equal to b,
          1 if a is greater than b
*/
static int compare_json_int_uint(longlong a, ulonglong b)
{
  // All negative values are less than the unsigned value b.
  if (a < 0)
    return -1;

  // If a is not negative, it is safe to cast it to ulonglong.
  return compare_numbers(static_cast<ulonglong>(a), b);
}


/**
  Compare the contents of two strings in a JSON value. The strings
  could be either JSON string scalars encoded in utf8mb4, or binary
  strings from JSON opaque scalars. In either case they are compared
  byte by byte.

  @param str1 the first string
  @param str1_len the length of str1
  @param str2 the second string
  @param str2_len the length of str2
  @retval -1 if str1 is less than str2,
  @retval 0 if str1 is equal to str2,
  @retval 1 if str1 is greater than str2
*/
static int compare_json_strings(const char *str1, size_t str1_len,
                                const char *str2, size_t str2_len)
{
  int cmp= memcmp(str1, str2, std::min(str1_len, str2_len));
  if (cmp != 0)
    return cmp;
  return compare_numbers(str1_len, str2_len);
}


/**
  The following matrix tells how two JSON values should be compared
  based on their types. If type_comparison[type_of_a][type_of_b] is
  -1, it means that a is smaller than b. If it is 1, it means that a
  is greater than b. If it is 0, it means it cannot be determined
  which value is the greater one just by looking at the types.
*/
static const int type_comparison[Json_dom::J_ERROR + 1][Json_dom::J_ERROR + 1]=
{
  /* NULL */      {0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* DECIMAL */   {1,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* INT */       {1,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* UINT */      {1,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* DOUBLE */    {1,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* STRING */    {1,  1,  1,  1,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* OBJECT */    {1,  1,  1,  1,  1,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1},
  /* ARRAY */     {1,  1,  1,  1,  1,  1,  1,  0, -1, -1, -1, -1, -1, -1, -1},
  /* BOOLEAN */   {1,  1,  1,  1,  1,  1,  1,  1,  0, -1, -1, -1, -1, -1, -1},
  /* DATE */      {1,  1,  1,  1,  1,  1,  1,  1,  1,  0, -1, -1, -1, -1, -1},
  /* TIME */      {1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0, -1, -1, -1, -1},
  /* DATETIME */  {1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0, -1, -1},
  /* TIMESTAMP */ {1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0, -1, -1},
  /* OPAQUE */    {1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0, -1},
  /* ERROR */     {1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1},
};


int Json_wrapper::compare(const Json_wrapper &other) const
{
  const Json_dom::enum_json_type this_type= type();
  const Json_dom::enum_json_type other_type= other.type();

  DBUG_ASSERT(this_type != Json_dom::J_ERROR);
  DBUG_ASSERT(other_type != Json_dom::J_ERROR);

  // Check if the type tells us which value is bigger.
  int cmp= type_comparison[this_type][other_type];
  if (cmp != 0)
    return cmp;

  // Same or similar type. Go on and inspect the values.

  switch (this_type)
  {
  case Json_dom::J_ARRAY:
    /*
      Two arrays are equal if they have the same length, and all
      elements in one array are equal to the corresponding elements in
      the other array.

      The array that has the smallest value on the first position that
      contains different values in the two arrays, is considered
      smaller than the other array. If the two arrays are of different
      size, and all values in the shorter array are equal to the
      corresponding values in the longer array, the shorter array is
      considered smaller.
    */
    {
      const size_t size_a= length();
      const size_t size_b= other.length();
      const size_t min_size= std::min(size_a, size_b);
      for (size_t i= 0; i < min_size; i++)
      {
        int cmp= (*this)[i].compare(other[i]);
        if (cmp != 0)
          return cmp;
      }
      return compare_numbers(size_a, size_b);
    }
  case Json_dom::J_OBJECT:
    /*
      An object is equal to another object if they have the same set
      of keys, and all values in one objects are equal to the values
      associated with the same key in the other object.
    */
    {
      /*
        If their sizes are different, the object with the smallest
        number of elements is smaller than the other object.
      */
      cmp= compare_numbers(length(), other.length());
      if (cmp != 0)
        return cmp;

      /*
        Otherwise, compare each key/value pair in the two objects.
        Return on the first difference that is found.
      */
      Json_wrapper_object_iterator it1= object_iterator();
      Json_wrapper_object_iterator it2= other.object_iterator();
      while (!it1.empty())
      {
        const std::pair<const std::string, Json_wrapper> elt1= it1.elt();
        const std::pair<const std::string, Json_wrapper> elt2= it2.elt();

        const std::string key1= elt1.first;
        const std::string key2= elt2.first;

        cmp= compare_json_strings(key1.data(), key1.size(),
                                  key2.data(), key2.size());
        if (cmp != 0)
          return cmp;

        cmp= elt1.second.compare(elt2.second);
        if (cmp != 0)
          return cmp;

        it1.next();
        it2.next();
      }

      DBUG_ASSERT(it1.empty());
      DBUG_ASSERT(it2.empty());

      // No differences found. The two objects must be equal.
      return 0;
    }
  case Json_dom::J_STRING:
    return compare_json_strings(get_data(), get_data_length(),
                                other.get_data(), other.get_data_length());
  case Json_dom::J_INT:
    // Signed integers can be compared to all other numbers.
    switch (other_type)
    {
    case Json_dom::J_INT:
      return compare_numbers(get_int(), other.get_int());
    case Json_dom::J_UINT:
      return compare_json_int_uint(get_int(), other.get_uint());
    case Json_dom::J_DOUBLE:
      return -compare_json_double_int(other.get_double(), get_int());
    case Json_dom::J_DECIMAL:
      {
        my_decimal a_dec;
        my_decimal b_dec;
        longlong2decimal(get_int(), &a_dec);
        if (other.get_decimal_data(&b_dec))
        {
          my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
          return 1;
        }
        return my_decimal_cmp(&a_dec, &b_dec);
      }
    default:
      break;
    }
  case Json_dom::J_UINT:
    // Unsigned integers can be compared to all other numbers.
    switch (other_type)
    {
    case Json_dom::J_UINT:
      return compare_numbers(get_uint(), other.get_uint());
    case Json_dom::J_INT:
      return -compare_json_int_uint(other.get_int(), get_uint());
    case Json_dom::J_DOUBLE:
      return -compare_json_double_uint(other.get_double(), get_uint());
    case Json_dom::J_DECIMAL:
       {
          my_decimal a_dec;
          my_decimal b_dec;
          ulonglong2decimal(get_uint(), &a_dec);
          if (other.get_decimal_data(&b_dec))
          {
            my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
            return 1;
          }
          return my_decimal_cmp(&a_dec, &b_dec);
        }
    default:
      break;
    }
  case Json_dom::J_DOUBLE:
    // Doubles can be compared to all other numbers.
    {
      switch (other_type)
      {
      case Json_dom::J_DOUBLE:
        return compare_numbers(get_double(), other.get_double());
      case Json_dom::J_INT:
        return compare_json_double_int(get_double(), other.get_int());
      case Json_dom::J_UINT:
        return compare_json_double_uint(get_double(), other.get_uint());
      case Json_dom::J_DECIMAL:
        {
          my_decimal other_dec;
          if (other.get_decimal_data(&other_dec))
          {
            my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
            return 1;
          }
          return -compare_json_decimal_double(other_dec, get_double());
        }
      default:
        break;
      }
    }
  case Json_dom::J_DECIMAL:
    // Decimals can be compared to all other numbers.
    {
      my_decimal a_dec;
      my_decimal b_dec;
      if (get_decimal_data(&a_dec))
      {
        my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
        return 1;
      }
      switch (other_type)
      {
      case Json_dom::J_DECIMAL:
        if (other.get_decimal_data(&b_dec))
        {
          my_error(ER_INVALID_JSON_BINARY_DATA, MYF(0));
          return 1;
        }
        return my_decimal_cmp(&a_dec, &b_dec);
      case Json_dom::J_INT:
        longlong2decimal(other.get_int(), &b_dec);
        return my_decimal_cmp(&a_dec, &b_dec);
      case Json_dom::J_UINT:
        ulonglong2decimal(other.get_uint(), &b_dec);
        return my_decimal_cmp(&a_dec, &b_dec);
      case Json_dom::J_DOUBLE:
        return compare_json_decimal_double(a_dec, other.get_double());
      default:
        break;
      }
    }
  case Json_dom::J_BOOLEAN:
    // Booleans are only equal to other booleans. false is less than true.
    return compare_numbers(get_boolean(), other.get_boolean());
  case Json_dom::J_DATETIME:
  case Json_dom::J_TIMESTAMP:
    // Timestamps and datetimes can be equal to each other.
    {
      MYSQL_TIME val_a;
      get_datetime(&val_a);
      MYSQL_TIME val_b;
      other.get_datetime(&val_b);
      return compare_numbers(TIME_to_longlong_packed(&val_a),
                             TIME_to_longlong_packed(&val_b));
    }
  case Json_dom::J_TIME:
  case Json_dom::J_DATE:
    // Dates and times can only be equal to values of the same type.
    {
      DBUG_ASSERT(this_type == other_type);
      MYSQL_TIME val_a;
      get_datetime(&val_a);
      MYSQL_TIME val_b;
      other.get_datetime(&val_b);
      return compare_numbers(TIME_to_longlong_packed(&val_a),
                             TIME_to_longlong_packed(&val_b));
    }
  case Json_dom::J_OPAQUE:
    /*
      Opaque values are equal to other opaque values with the same
      field type and the same binary representation.
    */
    cmp= compare_numbers(field_type(), other.field_type());
    if (cmp == 0)
      cmp= compare_json_strings(get_data(), other.get_data_length(),
                                other.get_data(), other.get_data_length());
    return cmp;
  case Json_dom::J_NULL:
    // Null is always equal to other nulls.
    DBUG_ASSERT(this_type == other_type);
    return 0;
  case Json_dom::J_ERROR:
    break;
  }

  DBUG_ABORT();                               /* purecov: inspected */
  return 1;                                   /* purecov: inspected */
}
